#!/usr/bin/env python3

'''
===============================================================================
This is the deformable mirror simulation module of XAOSIM.

It defines two classes of deformable mirrors:
- square grid membrane mirror (ALPAO or BMC membrane)
- segmented hexagonal mirror (BMC Hex)
===============================================================================
'''

import numpy as np
import threading
from .shmlib import shm
from . import pupil
import time
from PIL import Image


# ===========================================================
# ===========================================================
def influ_fun(iftype="cosine", sz=512, ifs=10):
    ''' -------------------------------------------------------------------
    Pre-computes an actuator influence function for a simulated setup.

    Parameters:
    ----------
    - iftype: influence function type (for now, only "cosine")
    - sz: size of the square array to compute (in pixels
    - ifs: influence function size (in pixels)

    Remarks:
    -------
    Anything else than "cos" for now will result in a "cone" shaped
    influence function. There is something that still needs to be adjusted
    here.
    ------------------------------------------------------------------- '''
    offset = 0.5 if (sz % 2 == 0) else 0.0

    xl = 1.0 * (np.arange(sz) - sz // 2 + offset)
    px = 0.5 * (1.0 + np.cos(np.pi*np.abs(xl)/ifs))

    dist = pupil._dist(sz, sz, between_pix=(sz % 2 == 0))
    if "cos" in iftype.lower():
        px = 0.5 * (1.0 + np.cos(np.pi*np.abs(xl)/ifs))
        px[np.abs(xl) > ifs] = 0.0
        res = np.outer(px, px)
    else:
        # if not cosine, then cone shaped!
        px = 1.0 - np.abs(xl) / ifs
        px[np.abs(xl) > ifs] = 0.0
        res = np.outer(px, px)
    res[dist > ifs] = 0.0
    return res


class DM(object):
    ''' -------------------------------------------------------------------
    Deformable mirror class

    The displacement maps *self.dmd* are in microns
    ------------------------------------------------------------------- '''

    # ==================================================
    def __init__(self, instrument="SCExAO", dms=50, nch=8,
                 shm_root="dmdisp", shdir="/dev/shm/",
                 csz=256, na0=50.0, dx=0.0, dy=0.0,
                 iftype="", ifr0=1.0, dtype=np.float64):
        ''' -----------------------------------------
        Constructor for instance of deformable mirror
        Parameters:
        ----------
        - instrument : a string
        - dms: an integer (linear size of the DM)
        - nch: number of channels
        - shm_root: the root name for shared mem files
        - shdir: directory where shared mem files will be

        Additions:
        ---------
        - csz: array size for computations
        - na0: number of actuators across
        - dx: DM L/R misalignment (in actuators)
        - dy: DM U/D misalignment (in actuators)

        Experimental:
        ------------
        - iftype: inf. function type ("cone", "cos", ...)
        - ifr0:   inf. function characteristic size (in actuator steps)
        ----------------------------------------- '''
        self.keepgoing = False
        self.dms = dms  # DM size of (dms x dms) actuators
        self.nch = nch  # numbers of channels to drive the DM
        self.dmtype = "square"  # square grid of actuators
        self.iftype = iftype    # type of influence function
        self.ifr0 = ifr0        # infl. function size (in actuators)
        self.dtype = dtype      # data type written to shared memory

        self.dmd0 = np.zeros((dms, dms), dtype=dtype)
        self.dmd = self.dmd0.copy()
        self.shm_cntr = np.zeros(nch) - 1

        self.shdir = shdir
        self.shm_root = shm_root
        self.shmf = shm_root + ".wf.shm"

        # --------------------------------------------------------------
        # create or overwrite shared memory data structures for the DM
        # --------------------------------------------------------------
        self.disp = shm(fname='%s%s.im.shm' % (shdir, shm_root),
                        data=self.dmd0, verbose=False)

        for i in range(nch):
            exec('''self.disp%d = shm(fname='%s%s%d.im.shm', 
                 data=self.dmd0, verbose=False)''' % (
                     i, shdir, shm_root, i))

        # -----------------------------------------------------
        # DM-induced wavefront shared memory data structure
        # -----------------------------------------------------
        self.wft = shm(fname='%s%s.wf.shm' % (shdir, shm_root),
                       data=np.zeros((csz, csz)), verbose=False)

        self.dx = dx
        self.dy = dy
        self.na0 = na0
        self.csz = csz
        self.astep = csz // na0
        # influence function characteristic and array sizes (in pixels)
        self._if_psz = self.astep*self.ifr0
        self._if_asz = int(np.round(2*self._if_psz))
        rwf = int(np.round(self.astep*self.dms))
        self.infun = influ_fun(iftype=self.iftype,
                               sz=self._if_asz, ifs=self._if_psz)

    # ==================================================
    def __str__(self):
        msg = f"Continuous Deformable Mirror:\n"
        msg += f"- {self.dms}x{self.dms} {self.dmtype} actuator grid\n"
        msg += f"- {self.nch} control channels\n"
        msg += f"- Computation size = {self.csz} "
        msg += f"({self.astep} pixels / actuators)\n"
        return msg

    # ==================================================
    def update_infun(self, iftype=None, ifr0=None):
        ''' ----------------------------------------
        Updates the DM influence function!

        Parameters:
        ----------
        - iftype: influence function type
        - ifr0: influence function radius (in actuators)
        ---------------------------------------- '''
        change = False
        if ifr0 is not None:
            self._if_psz = self.astep*self.ifr0
            self._if_asz = int(np.round(2*self._if_psz))
            change = True

        if iftype is not None:
            self.iftype = str(iftype)
            change = True

        if change is True:
            self.infun = influ_fun(
                iftype=self.iftype,
                sz=self._if_asz, ifs=self._if_psz)

    # ==================================================
    def close(self,):
        ''' ----------------------------------------
        Closes all the shared memory data structures
        for the DM (channels, global disp and wft)
        ---------------------------------------- '''
        self.disp.close()
        for ii in range(self.nch):
            exec('self.disp%d.close()' % (ii))
        self.wft.close()

    # ==================================================
    def get_counter_channel(self, chn=0):
        ''' ----------------------------------------
        Return the current channel counter value.
        Reads from the already-opened shared memory
        data structure.

        Parameters:
        ----------
        - chn  : the channel (integer)
        ---------------------------------------- '''
        cnt = 0
        if chn < self.nch:
            cnt = eval("self.disp%d.get_counter()" % (chn,))
        else:
            cnt = self.disp.get_counter()
        return(cnt)

    # ==================================================
    def reset_channel(self, chn=0):
        ''' ----------------------------------------
        Resets the state of the channel to zeros

        Parameters:
        ----------
        - chn  : the channel (integer)
        ---------------------------------------- '''
        if 0 <= chn < self.nch:
            eval("self.disp%d.set_data(self.dmd0)" % (chn,))

    # ==================================================
    def set_data_channel(self, dmap=None, chn=0):
        ''' ----------------------------------------
        Sends the dmap to the requested channel

        Parameters:
        ----------
        - dmap : float ndarray (size dms x dms)
        - chn  : the channel (integer)
        ---------------------------------------- '''
        if dmap is not None:
            if 0 <= chn < self.nch:
                eval("self.disp%d.set_data(dmap)" % (chn,))

    # ==================================================
    def get_data_channel(self, chn=0):
        ''' ----------------------------------------
        Returns the data present on a channel

        Parameters:
        ----------
        - chn: the channel (integer)
        ---------------------------------------- '''
        res = None
        if 0 <= chn < self.nch:
            res = eval("self.disp%d.get_data()" % (chn,))
        return res

    # ==================================================
    def start(self, delay=0.1):
        ''' ----------------------------------------
        Starts an independent thread that looks for
        changes on all channels, and updates the
        actual DM shape.
        ---------------------------------------- '''
        if not self.keepgoing:
            self.keepgoing = True
            t = threading.Thread(target=self.__loop__, args=(delay,))
            t.start()
            print("The *DM* server was started")
        else:
            print("The *DM* server is already running")

    # ==================================================
    def stop(self,):
        ''' ----------------------------------------
        Simple high-level accessor to interrupt the
        thread of the DM server infinite loop
        ---------------------------------------- '''
        if self.keepgoing:
            self.keepgoing = False
            print("The *DM* server was stopped")
        else:
            print("The *DM* server was not running")

    # ==================================================
    def update(self, verbose=False):
        ''' ----------------------------------------
        DM state update.

        Reads all existing channels and combines
        them to update the actual DM shape.
        ---------------------------------------- '''
        combi = np.zeros_like(self.disp0.get_data())
        for i in range(self.nch):
            exec("combi += self.disp%d.get_data()" % (i,))
        self.dmd = combi
        self.disp.set_data(combi)
        self.wft.set_data(self.map2D())
        if verbose:
            print("DM shape updated!")

    # ==================================================
    def __loop__(self, delay=0.1):
        ''' ----------------------------------------
        Thread (infinite loop) that updates the DM
        shape until told to stop.

        Do not use directly: call start()
        and stop() instead.
        ---------------------------------------- '''
        updt = True
        while self.keepgoing:
            for i in range(self.nch):
                test = self.get_counter_channel(i)
                if test != self.shm_cntr[i]:
                    self.shm_cntr[i] = test
                    updt = True
            if updt:
                updt = False
                self.update()
            time.sleep(delay)

    # ==================================================
    def map2D(self):
        ''' -----------------------------------------
        Returns a 2D displacement map of the DM.
        assuming a square grid of actuators

        Parameters:
        ----------
        - astep: actuator grid size (in pixels)
        ----------------------------------------- '''
        dmmap = np.zeros((self.csz, self.csz), dtype=np.float64)
        dms = self.dms
        dsz = self._if_asz  # padding for influence function
        astep = dsz // 2    # actuator step on map

        if self.iftype != "":  # influence function specified!
            amap = self.dmd.copy()

            na_off = (dms - self.na0) // 2  # offset actuators
            wsz = (dms+1) * astep          # work-map size
            wmap = np.zeros((wsz, wsz), dtype=np.float64)
            for jj in range(dms*dms):
                iy, ix = jj % dms, jj // dms     # actuator indices
                _y0 = int(np.round(iy*astep))
                _x0 = int(np.round(ix*astep))
                wmap[_y0:_y0+self._if_asz,
                     _x0:_x0+self._if_asz] += amap[iy, ix] * self.infun

            # keep useful part only
            y0 = int((na_off + 0.5 + self.dy) * astep)
            x0 = int((na_off + 0.5 + self.dx) * astep)
            dmmap = wmap[y0:y0+self.csz, x0:x0+self.csz]

        else:
            # interpolation using PIL!
            x0 = int(np.round((dsz - astep)/2))

            rwf = int(np.round(astep*dms))  # resized wavefront
            x0 = (self.csz-rwf) // 2
            x1 = x0 + rwf

            map0 = Image.fromarray(self.dmd)
            map1 = map0.resize((rwf, rwf), resample=1)
            dmmap[x0:x1, x0:x1] = map1

        return dmmap

# ===========================================================
# ===========================================================


class HexDM(DM):
    ''' -------------------------------------------------------------------
    Hexagonal Segmented Deformable mirror class

    ------------------------------------------------------------------- '''

    # ==================================================
    def __init__(self, instrument="KERNEL", nr=1, nch=8,
                 shm_root="hex_disp", shdir="/dev/shm/",
                 csz=512, na0=15.0, dx=0.0, dy=0.0,
                 srad=323.75):
        ''' -----------------------------------------
        Constructor for instance of deformable mirror
        Parameters:
        ----------
        - instrument : a string
        - dms: an integer (linear size of the DM)
        - nch: number of channels
        - shm_root: the root name for shared mem files

        Additions:
        ---------
        - csz: array size for computations
        - na0: number of segments across pupil diam
        - dx: DM L/R misalignment (in actuators)
        - dy: DM U/D misalignment (in actuators)
        - srad: segment radius (in microns)

        Note:
        ----
        Segment radius is defined as that of the circle
        that would be inscribed inside one segment.
        ----------------------------------------- '''
        self.keepgoing = False
        self.dmtype = "hex"
        self.nr = nr             # DM number of rings
        self.ns = 1+3*nr*(nr+1)  # DM number of segments
        self.nch = nch           # numbers of DM channels
        self.dmd0 = np.zeros((self.ns, 3), dtype=np.float32)
        self.shm_cntr = np.zeros(nch) - 1
        self.disp = shm(fname='%s%s.im.shm' % (shdir, shm_root),
                        data=self.dmd0, verbose=False, nbkw=1)

        self.shdir = shdir
        self.shm_root = shm_root
        self.shmf = shm_root + ".wf.shm"

        # share dmtype with instrument via keywords
        self.disp.update_keyword(0, "dmtype", self.dmtype,
                                 "geometry of the Deformable Mirror")

        for ii in range(nch):
            exec('''self.disp%d = shm(fname='%s%s%d.im.shm', 
            data=self.dmd0, verbose=False)''' % (
                ii, shdir, shm_root, ii))

        # additional shared memory data structure for the wavefront
        self.wft = shm(fname='%s%s.wf.shm' % (shdir, shm_root),
                       data=np.zeros((csz, csz)), verbose=False)
        self.dx = dx
        self.dy = dy
        self.na0 = na0
        self.csz = csz
        self.astep = csz // na0
        self.srad = srad

    # ==================================================
    def map2D(self):
        ''' -----------------------------------------
        Returns a 2D displacement map of the Hex DM.

        Parameters:
        ----------
        - astep : actuator pitch (in pixels)
        ----------------------------------------- '''
        nr = self.nr
        msz = self.csz
        arad = self.astep/np.sqrt(3)
        dmmap = np.zeros((msz, msz), dtype=np.float64)
        xx, yy = np.meshgrid(np.arange(msz)-msz/2, np.arange(msz)-msz/2)

        scoeff = 2*self.srad*1e-3/self.astep  # slope coefficient

        ttx = self.dmd[:, 1] / scoeff
        tty = self.dmd[:, 2] / scoeff

        xy = pupil.hex_grid_coords(nr+1, astep, rot=2*np.pi/3)
        xy[0] += self.dx * astep  # optional offset of
        xy[1] += self.dy * astep  # the DM position
        xy = np.round(xy).astype(np.int)

        seg = pupil.uniform_hex(msz, msz, arad).T

        for ii in range(self.ns):
            seg1 = seg*(self.dmd[ii, 0]*seg + ttx[ii]*xx + tty[ii]*yy)
            seg1 = np.roll(np.roll(seg1, xy[0, ii], axis=0), xy[1, ii], axis=1)
            dmmap += seg1

        return dmmap

    # ==================================================
    def disp_2_PTT(self, dmap=None):
        ''' -------------------------------------------------------------------
        Turns the provided DM displacement map into piston-tip-tilt commands
        for the HexDM

        Parameters:
        ----------
        - dmap: displacement map the HexDM will attempt to approximate
                 expressed in microns

        Returns: a (nseg x 3) numpy array that contains:
        -------
        - the piston in the first column (in microns)
        - the x-tip  in the second column (in mrad)
        - the y-tilt in the third column (in mrad)

        Remarks:
        -------

        To convert a desired wavefront (in radians), into displacement
        map on the DM (in microns), things must be scaled by the wavelength.
        Just make sure to also take the x2 effect the DM displacement has
        on the wavefront.

        The required scaling parameter will typically be *lambda / (4*PI)*
        ------------------------------------------------------------------- '''
        nr = self.nr
        arad = self.astep/np.sqrt(3)

        xx, yy = np.meshgrid(np.arange(self.csz)-self.csz/2,
                             np.arange(self.csz)-self.csz/2)

        xy = pupil.hex_grid_coords(nr+1, self.astep, rot=2*np.pi/3)
        xy[0] += self.dx * self.astep  # optional offset of
        xy[1] += self.dy * self.astep  # the DM position
        xy = np.round(xy).astype(np.int)

        scoeff = 2*self.srad*1e-3/self.astep  # slope coefficient

        # centered reference segment
        seg = pupil.uniform_hex(self.csz, self.csz, arad).T

        dmd0 = np.zeros_like(self.dmd0)  # empty tip-tilt-piston commands

        xxnorm = xx[seg > 0].dot(xx[seg > 0])
        yynorm = yy[seg > 0].dot(yy[seg > 0])

        for ii in range(self.ns):
            tmp = np.roll(np.roll(dmap, -xy[0, ii], axis=0),
                          -xy[1, ii], axis=1)
            dmd0[ii, 0] = tmp[seg > 0].mean()
            dmd0[ii, 1] = scoeff * tmp[seg > 0].dot(xx[seg > 0]) / xxnorm
            dmd0[ii, 2] = scoeff * tmp[seg > 0].dot(yy[seg > 0]) / yynorm

        return dmd0

    # ========================================================================
    def ptt_2_actuator(self, ptt, piston_only=True):
        ''' -------------------------------------------------------------------
        converts an array of piston+tip-tilt segment states and converts them
        into actuator commands for the Hex BMC mirror.

        Parameters:
        ----------
        - ptt:         the array of piston+tip-tilt command for all segments
        - piston_only: option to apply piston correction only (easier scenario)

        Remarks:
        -------
        - #1 Assumes that the ptt commands are expressed in:
        - microns (for the piston)
        - mrad (for the tip-tilt)
        - #2 Commands sent to the DM are floating point numbers 0 <= cmd <= 1.
        ------------------------------------------------------------------- '''
        csz = 1024   # command size
        nseg = 169   # number of segments
        again = 4.0  # actuator gain: 4 um per ADU
        cmd = np.zeros(csz)

        a0 = self.srad  # 323.75 um for the BMC Hex DM
        mat = np.array([[1,                0,  a0],
                        [1, -np.sqrt(3)/2*a0, -a0/2],
                        [1,  np.sqrt(3)/2*a0, -a0/2]])

        # for the piston only scenario
        if piston_only:
            for ii in range(nseg):
                cmd[ii*3] = ptt[ii, 0]
                cmd[ii*3+1] = ptt[ii, 0]
                cmd[ii*3+2] = ptt[ii, 0]

        else:
            print("TBD!!")

        cmd /= again  # command converted into BMC HexDM ADU

        return cmd



#!/usr/bin/env python3
''' ===================================================================
Test dodumentation for SFT, the "slow" Fourier Transform using the
theory described in: http://adsabs.harvard.edu/abs/2007OExpr..1515935S

Example:
-------

>> import pupil
>> import sft
>> a = pupil.uniform_disk(200,200, 100)
>> ca_focal = sft.sft(a, 50, 20)

Will compute the complex amplitude in the focal plane for a pupil
over a +/- 10 lambda/D field of view

>> lyot = sft.isft(ca_focal, 200, 20)

Will compute the complex amplitude in a pupil plane located downstream
from the previous focal plane, using only the spatial frequencies
present in the array that was previously computed.
'''

import numpy as np
i2pi = 1j*2*np.pi


# ===================================================================
def sft(A2, NB, m, inv=False, btwn_pix=False):
    ''' --------------------------------------------------------------
    Explicit Fourier Transform, using the theory described in:
    http://adsabs.harvard.edu/abs/2007OExpr..1515935S

    Assumes the original array is square.
    No need to "center" the data on the origin.

    Parameters:
    ----------

    - A2 : the 2D original array
    - NB : the linear size of the result array (integer)
    - m  : m/2 = maximum spatial frequency to be computed (in l/D)
    - inv: boolean (direct or inverse) see the definition of isft()
    - btwn_pix: if True, computation centered between pixels

    Returns the Fourier transform of A2 over an array of size NBxNB
    -------------------------------------------------------------- '''

    offset = 0
    if btwn_pix is True:
        offset = 0.5

    NA = np.shape(A2)[0]
    m = float(m)
    coeff = m/(NA*NB)

    U = np.zeros((1, NB))
    X = np.zeros((1, NA))

    X[0, :] = (1./NA)*(np.arange(NA)-NA/2.0+offset)
    U[0, :] = (m/NB)*(np.arange(NB)-NB/2.0+offset)

    sign = -1.0
    if inv:
        sign = 1.0

    A1 = np.exp(sign*i2pi*np.dot(np.transpose(U), X))
    A3 = np.exp(sign*i2pi*np.dot(np.transpose(X), U))

    B = A1.dot(A2.dot(A3))
    return coeff*np.array(B)


# ===================================================================
def isft(A2, NB, m):
    ''' --------------------------------------------------------------
    Explicit inverse Fourier Transform, using the theory described in:
    http://adsabs.harvard.edu/abs/2007OExpr..1515935S

    See documentation for sft().
    -------------------------------------------------------------- '''
    return sft(A2, NB, m, inv=True)



#!/usr/bin/env python3

'''---------------------------------------------------------------------------
Read and write access to shared memory (SHM) structures used by SCExAO

- Author : Frantz Martinache
- Date   : June 13, 2018
- Revised: May 7, 2019

Improved version of the original SHM structure used by SCExAO and friends.
---------------------------------------------------------------------------

Semaphores required by SCExAO call for the specialized scexao_shm module.
This module relies on the posix_ipc library available on PyPi

---------------------------------------------------------------------------
Note on data alignment (refer to section 7.3.2.1 of python documentation)

By default, C types are represented in the machine's native format and byte
order, and properly aligned by skipping pad bytes if necessary (according to
the rules used by the C compiler).

To request no alignment, while using native byte-order, the first character
of the format string must be "="! This is used for keywords.

Between 2017 and 2018, SCExAO began adopting an aligned data-structure
again. This class was modified accordingly and includes a "packed"
constructor option to choose between the two styles.

The packed mode was eventually discarded. It is still available as an
option here but is not recommended and will eventually be deprecated.
---------------------------------------------------------------------------

Note on the order of axes for the data:
The convention for the shared memory structure follows that of the fits file:
for a 3D array, the axes are in the x,y,z order.

Internally in this library, the axes are reversed: z,y,x.
---------------------------------------------------------------------------
'''

import os
import mmap
import struct
import numpy as np
import time

# ------------------------------------------------------
#          list of available data types
# ------------------------------------------------------
all_dtypes = [np.uint8,     np.int8,    np.uint16,    np.int16,
              np.uint32,    np.int32,   np.uint64,    np.int64,
              np.float32,   np.float64, np.complex64, np.complex128]

# # ------------------------------------------------------
# #    string used to decode the binary shm structure
# # ------------------------------------------------------
hdr_fmt_pck = '32s 80s B   3I Q B   Q 2Q 2Q 2Q 2Q   B b B   Q   B   H   Q Q Q B   H 64s' # packed style
hdr_fmt_aln = '32s 80s B3x 3I Q B7x Q 2Q 2Q 2Q 2Q   B b B5x Q   B1x H4x Q Q Q B1x H 64s4x' # aligned style
c0_hdr_pos = 17  # position of the counter #0 in the header.

# ------------------------------------------------------
# list of metadata keys for the shm structure (global)
# these keys have to match the header introduced above!
# ------------------------------------------------------
mtkeys = ['bversion', 'bimname',
          'naxis', 'x', 'y', 'z', 'nel', 'atype', 'imtype',
          'crtime_sec', 'crtime_ns',
          'latime_sec', 'latime_ns',
          'atime_sec', 'atime_ns',
          'wtime_sec', 'wtime_ns',
          'shared', 'loc', 'status',
          'flag', 'logflag', 'sem',
          'cnt0', 'cnt1', 'cnt2',
          'write', 'nbkw', 'bcudamem']

'''

One of the most important changes when moving from python2
to python3 for this library was that the string C type
described by the format code "s" went from python data
type "str" to "bytes", likely to  acommodate unicode.

---------------------------------------------------------
Table taken from Python 3 documentation, section 7.1.2.2.
---------------------------------------------------------

|--------+--------------------+---------------+----------|
| Format | C Type             | Python type   | Std size |
|--------+--------------------+---------------+----------|
| x      | pad byte           | no value      |          |
| c      | char               | bytes (len=1) |        1 |
| b      | signed char        | integer       |        1 |
| B      | unsigned char      | integer       |        1 |
| ?      | _Bool              | bool          |        1 |
| h      | short              | integer       |        2 |
| H      | unsigned short     | integer       |        2 |
| i      | int                | integer       |        4 |
| I      | unsigned int       | integer       |        4 |
| l      | long               | integer       |        4 |
| L      | unsigned long      | integer       |        4 |
| q      | long long          | integer       |        8 |
| Q      | unsigned long long | integer       |        8 |
| n      | ssize_t            | integer       |          |
| N      | size_t             | integer       |          |
| e      | (7)                | float         |        2 |
| f      | float              | float         |        4 |
| d      | double             | float         |        8 |
| s      | char[]             | bytes         |          |
| p      | char[]             | bytes         |          |
| P      | void *             | integer       |          |
|--------+--------------------+---------------+----------|

'''


class shm:
    def __init__(self, fname=None, data=None, verbose=False,
                 packed=False, nbkw=0):
        ''' --------------------------------------------------------------
        Constructor for a SHM (shared memory) object.

        Parameters:
        ----------
        - fname: name of the shared memory file structure
        - data: some array (1, 2 or 3D of data)
        - verbose: optional boolean
        - packed: True -> packed / False -> aligned data format
        - nbkw: # of keywords to be appended to the data structure (optional)

        Depending on whether the file already exists, and/or some new
        data is provided, the file will be created or overwritten.
        -------------------------------------------------------------- '''
        self.packed = packed

        if self.packed:
            self.hdr_fmt = hdr_fmt_pck  # packed shm structure
            self.kwfmt0 = "16s s"       # packed keyword structure
        else:
            self.hdr_fmt = hdr_fmt_aln  # aligned shm structure
            self.kwfmt0 = "16s s7x"     # aligned keyword structure

        self.c0_offset = 0  # fast-offset for counter #0 (updated later)
        self.kwsz = struct.calcsize('16s 80s'+' '+self.kwfmt0)  # keyword size

        self.empty = False  # control variable

        # --------------------------------------------------------------------
        #                dictionary containing the metadata
        # --------------------------------------------------------------------

        self.mtdata = {
            'version': '', 'bversion': b'',
            'imname': '', 'bimname': b'',
            'naxis': 0, 'x': 0, 'y': 0, 'z': 0,
            'size': (0, 0, 0), 'nel': 0, 'atype': 0, 'imtype': 0,
            'crtime_sec': 0, 'crtime_ns': 0, 'crtime': (0, 0),
            'latime_sec': 0, 'latime_ns': 0, 'latime': (0, 0),
            'atime_sec': 0,  'atime_ns': 0,  'atime': (0, 0), 'atimearr': 0,
            'wtime_sec': 0,  'wtime_ns': 0,  'wtime': (0, 0), 'wtimearr': 0,
            'shared': 0, 'loc': 0, 'status': 0, 'flag': 0,
            'flagarr': 0, 'logflag': 0, 'sem': 0,
            'cnt0': 0, 'cnt1': 0, 'cnt2': 0, 'cntarr': 0,
            'write': 0, 'nbkw': 0,
            'cudamem': '', 'bcudamem': b''}

        # --------------------------------------------------------------------
        #          dictionary describing the content of a keyword
        # --------------------------------------------------------------------
        self.kwd = {'name': '', 'type': 'N', 'value': '', 'comment': ''}

        fmt = self.hdr_fmt
        self.c0_offset = struct.calcsize(' '.join(fmt.split()[:c0_hdr_pos]))
        self.im_offset = struct.calcsize(fmt)

        # ---------------
        if fname is None:
            print("No SHM file name provided")
            return(None)

        # ---------------
        self.fname = fname
        if data is not None:
            print("%s will be overwritten" % (fname,))
            self.create(fname, data, nbkw)

        else:
            if verbose:
                print("reading from existing %s" % (fname,))

            if not os.path.exists(fname):
                self.empty = True
                return

            self.fd = os.open(fname, os.O_RDWR)
            self.stats = os.fstat(self.fd)
            self.buf_len = self.stats.st_size
            self.buf = mmap.mmap(self.fd, self.buf_len, mmap.MAP_SHARED)
            self.read_meta_data(verbose=verbose)

            self.select_dtype()         # identify main data-type
            self.get_data()             # read the main data
            self.create_keyword_list()  # create empty list of keywords
            self.read_keywords()        # populate the keywords with data

    def create(self, fname, data, nbkw=0):
        ''' --------------------------------------------------------------
        Create a shared memory data structure

        Parameters:
        ----------
        - fname: name of the shared memory file structure
        - data: some array (1, 2 or 3D of data)

        Called by the constructor if the provided file-name does not
        exist: a new structure needs to be created, and will be populated
        with information based on the provided data.
        -------------------------------------------------------------- '''

        if data is None:
            print("No data (ndarray) provided! Nothing happens here")
            return

        # ---------------------------------------------------------
        # feed the relevant dictionary entries with available data
        # ---------------------------------------------------------
        self.npdtype = data.dtype
        self.mtdata['imname'] = fname.ljust(80, ' ')
        self.mtdata['bimname'] = bytes(self.mtdata['imname'], 'ascii')

        self.mtdata['version'] = "xaosim".ljust(32, ' ')
        self.mtdata['bversion'] = bytes(self.mtdata['version'], 'ascii')

        self.mtdata['naxis'] = data.ndim
        self.mtdata['size'] = data.shape[:data.ndim][::-1]
        self.mtdata['nel'] = data.size
        self.mtdata['atype'] = self.select_atype()
        self.mtdata['shared'] = 1
        self.mtdata['nbkw'] = nbkw

        if data.ndim == 3:
            self.mtdata['x'] = self.mtdata['size'][0]
            self.mtdata['y'] = self.mtdata['size'][1]
            self.mtdata['z'] = self.mtdata['size'][2]
        if data.ndim == 2:
            self.mtdata['x'] = self.mtdata['size'][0]
            self.mtdata['y'] = self.mtdata['size'][1]
            self.mtdata['size'] = self.mtdata['size'] + (0,)

        self.select_dtype()

        # ---------------------------------------------------------
        #          reconstruct a SHM metadata buffer
        # ---------------------------------------------------------
        fmt = self.hdr_fmt
        temp = [self.mtdata[mtkeys[ii]] for ii in range(len(mtkeys))]
        minibuf = struct.pack(fmt, *temp)

        # ---------------------------------------------------------
        #             allocate the file and mmap it
        # ---------------------------------------------------------
        kwspace = self.kwsz * nbkw                     # kword space
        fsz = self.im_offset + self.img_len + kwspace  # file size
        npg = fsz // mmap.PAGESIZE + 1                 # nb pages

        self.fd = os.open(fname, os.O_CREAT | os.O_TRUNC | os.O_RDWR)
        os.fchmod(self.fd, 0o777)  # give RWX access to all users
        os.write(self.fd, b'\x00' * npg * mmap.PAGESIZE)
        self.buf = mmap.mmap(self.fd, npg * mmap.PAGESIZE,
                             mmap.MAP_SHARED, mmap.PROT_WRITE)

        # ---------------------------------------------------------
        #              write the information to SHM
        # ---------------------------------------------------------
        self.buf[:self.im_offset] = minibuf  # the metadata

        self.set_data(data)
        self.create_keyword_list()
        self.write_keywords()

    def rename_img(self, newname):
        ''' --------------------------------------------------------------
        Gives the user a chance to rename the image.

        Parameter:
        ---------
        - newname: a string (< 80 char) with the name
        -------------------------------------------------------------- '''

        self.mtdata['imname'] = newname.ljust(80, ' ')
        self.mtdata['bimname'] = bytes(self.mtdata['imname'], 'ascii')
        self.buf[0:80] = struct.pack('80s', self.mtdata['bimname'])

    def close(self, erase_file=True):
        ''' --------------------------------------------------------------
        Clean close of a SHM data structure link

        - Clean close of buffer
        - release the file descriptor
        - delete shared memory file
        -------------------------------------------------------------- '''
        c0 = self.c0_offset                      # counter offset
        # self.buf[c0:c0+8] = struct.pack('Q', 0)  # set counter to zero
        self.buf.close()
        os.close(self.fd)
        self.fd = 0
        if erase_file:
            try:
                os.remove(self.fname)
                print("%s erased!" % (self.fname,))
            except FileNotFoundError:
                print("%s no longer exists already?" % (self.fname,))
        return(0)

    def read_meta_data(self, verbose=True):
        ''' --------------------------------------------------------------
        Read the metadata fraction of the SHM file.
        Populate the shm object mtdata dictionary.

        Parameters:
        ----------
        - verbose: (boolean, default: True), prints its findings
        -------------------------------------------------------------- '''
        fmt = self.hdr_fmt
        hlen = struct.calcsize(fmt)
        temp = struct.unpack(fmt, self.buf[:hlen])

        for ii in range(len(mtkeys)):
            self.mtdata[mtkeys[ii]] = temp[ii]

        # special repackaging: image name (string) and size (tuple)
        self.mtdata['imname'] = self.mtdata['bimname'].decode('ascii').strip('\x00')
        self.mtdata['version'] = self.mtdata['bversion'].decode('ascii').strip('\x00')
        self.mtdata['cudamem'] = self.mtdata['bcudamem'].decode('ascii').strip('\x00')
        self.mtdata['size'] = self.mtdata['z'], self.mtdata['y'], self.mtdata['x']
        self.mtdata['crtime'] = self.mtdata['crtime_sec'], self.mtdata['crtime_ns']
        self.mtdata['latime'] = self.mtdata['latime_sec'], self.mtdata['latime_ns']
        self.mtdata['atime'] = self.mtdata['atime_sec'], self.mtdata['atime_ns']
        self.mtdata['wtime'] = self.mtdata['wtime_sec'], self.mtdata['wtime_ns']

        if verbose:
            self.print_meta_data()

    def create_keyword_list(self):
        ''' --------------------------------------------------------------
        Place-holder. The name should be sufficiently explicit.
        -------------------------------------------------------------- '''
        nbkw = self.mtdata['nbkw']     # how many keywords
        self.kwds = []                 # prepare an empty list
        for ii in range(nbkw):         # fill with empty dictionaries
            self.kwds.append(self.kwd.copy())

    def read_keywords(self):
        ''' --------------------------------------------------------------
        Read all keywords from SHM file
        -------------------------------------------------------------- '''
        for ii in range(self.mtdata['nbkw']):
            self.read_keyword(ii)

    def write_keywords(self):
        ''' --------------------------------------------------------------
        Writes all keyword data to SHM file
        -------------------------------------------------------------- '''
        for ii in range(self.mtdata['nbkw']):
            self.write_keyword(ii)

    def read_keyword(self, ii):
        ''' --------------------------------------------------------------
        Read the content of keyword of given index.

        Parameters:
        ----------
        - ii: index of the keyword to read
        -------------------------------------------------------------- '''
        kwsz = self.kwsz              # keyword SHM data structure size
        k0 = self.im_offset + self.img_len + ii * kwsz  # kword offset

        # ------------------------------------------
        #             read from SHM
        # ------------------------------------------
        kwlen = struct.calcsize(self.kwfmt0)
        kname, ktype = struct.unpack(self.kwfmt0, self.buf[k0:k0+kwlen])

        # ------------------------------------------
        # depending on type, select parsing strategy
        # ------------------------------------------
        kwfmt = '16s 80s'

        if ktype == b'L':    # keyword value is int64
            kwfmt = 'q 8x 80s'
        elif ktype == b'D':  # keyword value is double
            kwfmt = 'd 8x 80s'
        elif ktype == b'S':  # keyword value is string
            kwfmt = '16s 80s'
        elif ktype == b'N':  # keyword is unused
            kwfmt = '16s 80s'

        kval, kcomm = struct.unpack(kwfmt, self.buf[k0+kwlen:k0+kwsz])

        if kwfmt == '16s 80s':
            kval = str(kval).strip('\x00')

        # ------------------------------------------
        #    fill in the dictionary of keywords
        # ------------------------------------------
        if (ktype == b'L'):
            self.kwds[ii]['value'] = np.long(kval)
        elif (ktype == b'D'):
            self.kwds[ii]['value'] = np.double(kval)
        else:
            self.kwds[ii]['value'] = ktype.decode('ascii').strip('\x00')

        self.kwds[ii]['name'] = kname.decode('ascii').strip('\x00')
        self.kwds[ii]['type'] = ktype.decode('ascii')
        self.kwds[ii]['comment'] = kcomm.decode('ascii').strip('\x00')

    def update_keyword(self, ii, name, value, comment):
        ''' --------------------------------------------------------------
        Update keyword data in dictionary and writes it to SHM file

        Parameters:
        ----------
        - ii      : index of the keyword to write (integer)
        - name    : the new keyword name
        -------------------------------------------------------------- '''

        if (ii >= self.mtdata['nbkw']):
            print("Keyword index %d is not allocated and cannot be written")
            return

        # ------------------------------------------
        #    update relevant keyword dictionary
        # ------------------------------------------
        try:
            self.kwds[ii]['name'] = str(name).ljust(16, ' ')
        except:
            print('Keyword name not compatible (< 16 char)')

        if isinstance(value, int):
            self.kwds[ii]['type'] = 'L'
            self.kwds[ii]['value'] = np.long(value)

        elif isinstance(value, float):
            self.kwds[ii]['type'] = 'D'
            self.kwds[ii]['value'] = np.double(value)

        elif isinstance(value, str):
            self.kwds[ii]['type'] = 'S'
            self.kwds[ii]['value'] = str(value)
        else:
            self.kwds[ii]['type'] = 'N'
            self.kwds[ii]['value'] = str(value)

        try:
            self.kwds[ii]['comment'] = str(comment).ljust(80, ' ')
        except:
            print('Keyword comment not compatible (< 80 char)')

        # ------------------------------------------
        #          write keyword to SHM
        # ------------------------------------------
        self.write_keyword(ii)

    def write_keyword(self, ii):
        ''' --------------------------------------------------------------
        Write keyword data to shared memory.

        Parameters:
        ----------
        - ii      : index of the keyword to write (integer)
        -------------------------------------------------------------- '''

        if (ii >= self.mtdata['nbkw']):
            print("Keyword index %d is not allocated and cannot be written")
            return

        kwsz = self.kwsz
        k0 = self.im_offset + self.img_len + ii * kwsz  # kword offset

        # ------------------------------------------
        #    read the keyword dictionary
        # ------------------------------------------
        kname = self.kwds[ii]['name']
        ktype = self.kwds[ii]['type']
        kval = self.kwds[ii]['value']
        kcomm = self.kwds[ii]['comment']

        if ktype == 'L' or ktype == b'L':
            kwfmt = '='+self.kwfmt0+' q 8x 80s'
            tmp = (bytes(kname, "ascii"),
                   bytes(ktype, "ascii"),
                   kval,
                   bytes(kcomm, "ascii"))
        elif ktype == 'D' or ktype == b'D':
            kwfmt = '='+self.kwfmt0+' d 8x 80s'
            tmp = (bytes(kname, "ascii"),
                   bytes(ktype, "ascii"),
                   kval,
                   bytes(kcomm, "ascii"))
        else:  # 'S' or 'N'
            kwfmt = '='+self.kwfmt0+' 16s 80s'
            tmp = (bytes(kname, "ascii"),
                   bytes(ktype, "ascii"),
                   bytes(kval, "ascii"),
                   bytes(kcomm, "ascii"))

        self.buf[k0:k0+kwsz] = struct.pack(kwfmt,  *tmp)

    def print_meta_data(self):
        ''' --------------------------------------------------------------
        Basic printout of the content of the mtdata dictionary.
        -------------------------------------------------------------- '''
        for ii in range(len(mtkeys)):
            print(mtkeys[ii], self.mtdata[mtkeys[ii]])

    def select_dtype(self):
        ''' --------------------------------------------------------------
        Based on the value of the 'atype' code used in SHM, determines
        which numpy data format to use.
        -------------------------------------------------------------- '''
        atype = self.mtdata['atype']
        self.npdtype = all_dtypes[atype-1]
        self.img_len = self.mtdata['nel'] * self.npdtype().itemsize

    def select_atype(self):
        ''' --------------------------------------------------------------
        Based on the type of numpy data provided, sets the appropriate
        'atype' value in the metadata of the SHM file.
        -------------------------------------------------------------- '''
        for i, mydt in enumerate(all_dtypes):
            if mydt == self.npdtype:
                self.mtdata['atype'] = i+1
        return(self.mtdata['atype'])

    def get_counter(self,):
        ''' --------------------------------------------------------------
        Read the image counter from SHM
        -------------------------------------------------------------- '''
        c0 = self.c0_offset                              # counter offset
        cntr = struct.unpack('Q', self.buf[c0:c0+8])[0]  # read from SHM
        self.mtdata['cnt0'] = cntr                       # update object mtdata
        return(cntr)

    def increment_counter(self,):
        ''' --------------------------------------------------------------
        Increment the image counter. Called when writing new data to SHM
        -------------------------------------------------------------- '''
        c0 = self.c0_offset                         # counter offset
        cntr = self.get_counter() + 1               # increment counter
        self.buf[c0:c0+8] = struct.pack('Q', cntr)  # update SHM file
        self.mtdata['cnt0'] = cntr                  # update object mtdata
        return(cntr)

    def get_data(self, check=False, reform=True, sleepT=0.001, timeout=5):
        ''' --------------------------------------------------------------
        Reads and returns the data part of the SHM file

        Parameters:
        ----------
        - check: integer (last index) if not False, waits image update
        - reform: boolean, if True, reshapes the array in a 2-3D format
        - sleepT: time increment (in seconds) when waiting for new data
        - timeout: timeout in seconds
        -------------------------------------------------------------- '''
        i0 = self.im_offset     # image offset
        i1 = i0 + self.img_len  # image end

        time0 = time.time()
        if check is not False:
            timen = time.time()

            while ((self.get_counter() <= check) and (timen-time0 < timeout)):
                time.sleep(sleepT)
                timen = time.time()

        data = np.fromstring(self.buf[i0:i1], dtype=self.npdtype)

        if reform:
            if self.mtdata['naxis'] == 2:
                rsz = self.mtdata['y'], self.mtdata['x']
            else:
                rsz = self.mtdata['z'], self.mtdata['y'], self.mtdata['x']
            data = np.reshape(data, rsz)
        return(data)

    def set_data(self, data, check_dt=False):
        ''' --------------------------------------------------------------
        Upload new data to the SHM file.

        Parameters:
        ----------
        - data: the array to upload to SHM
        - check_dt: boolean (default: false) recasts data

        Note:
        ----

        The check_dt is available here for comfort. For the sake of
        performance, data should be properly cast to start with, and
        this option not used!
        -------------------------------------------------------------- '''
        i0 = self.im_offset     # image offset
        i1 = i0 + self.img_len  # image end

        if check_dt is True:
            self.buf[i0:i1] = data.astype(self.npdtype()).tostring()
        else:
            try:
                self.buf[i0:i1] = data.tostring()
            except:
                print("Warning: writing wrong data-type to shared memory")
                return
        self.increment_counter()

        return

    def save_as_fits(self, fitsname):
        ''' --------------------------------------------------------------
        Convenient sometimes, to be able to export the data as a fits file.
        Parameters:
        ----------
        - fitsname: a filename (clobber=True)
        -------------------------------------------------------------- '''
        try:
            import astropy.io.fits as pf
            pf.writeto(fitsname, self.get_data(), overwrite=True)
        except ModuleNotFoundError:
            import pyfits as pf
            pf.writeto(fitsname, self.get_data(), clobber=True)
        return(0)

# =================================================================
# =================================================================



#!/usr/bin/env python

'''---------------------------------------------------------------------------
Read and write access to shared memory (SHM) structures used by SCExAO

- Author : Frantz Martinache
- Date   : June 13, 2018

Improved version of the original SHM structure used by SCExAO and friends.

Inherits from the more generic shm class defined in the shmlib module,
and includes additional functions expecting specific keywords to be
implemented, when dealing with camera images.
---------------------------------------------------------------------------

'''

from .shmlib import shm as shm0
import numpy as np
import posix_ipc as ipc
import os


class shm(shm0):
    # =====================================================================
    def __init__(self, fname=None, data=None,
                 verbose=False, nbkw=0):
        ''' --------------------------------------------------------------
        Constructor for a SHM (shared memory) object.

        Parameters:
        ----------
        - fname: name of the shared memory file structure
        - data: some array (1, 2 or 3D of data)
        - verbose: optional boolean
        - nbkw: # of keywords to be appended to the data structure (optional)

        Depending on whether the file already exists, and/or some new
        data is provided, the file will be created or overwritten.

        In addition to the generic shm data structure, semaphores are
        connected to the object.
        -------------------------------------------------------------- '''
        self.nosem = True
        shm0.__init__(self, fname, data, verbose, False, nbkw)
        self.nsem = 10  # number of semaphores to address
        myname = os.path.basename(fname).split('.')[0]

        for ii in range(self.nsem):
            semf = "%s_sem%02d" % (myname, ii)
            exec('self.sem%02d = ipc.Semaphore(semf, ipc.O_RDWR | ipc.O_CREAT)' % (ii,))
            _ = ipc.Semaphore("%s_semlog" % (myname,),
                              ipc.O_RDWR | ipc.O_CREAT)
        self.nosem = False

    # =====================================================================
    def set_data(self, data, check_dt=False):
        ''' --------------------------------------------------------------
        On SCExAO, in addition to updating the actual DM data, on must post
        semaphores to signal the DM to update
        -------------------------------------------------------------- '''
        shm0.set_data(self, data, check_dt)
        if self.nosem is False:
            for ii in range(10):
                semf = "%s_sem%02d" % (self.mtdata['imname'], ii)
                exec('self.sem%02d.release()' % (ii,))
        else:
            print("skip sem post this first time")

    # =====================================================================
    def close(self,):
        shm0.close(self)
        for ii in range(self.nsem):
            semf = "%s_sem%02d" % (self.mtdata['imname'], ii)
            exec('self.sem%02d.close()' % (ii,))

    # =====================================================================
    def get_expt(self,):
        ''' --------------------------------------------------------------
        SCExAO specific: returns the exposure time (from keyword)
        -------------------------------------------------------------- '''
        ii0 = 0  # index of exposure time in keywords
        self.read_keyword(ii0)
        self.expt = self.kwds[ii0]['value']
        return self.expt

    # =====================================================================
    def get_fps(self,):
        ''' --------------------------------------------------------------
        SCExAO specific: returns the frame rate (from keyword)
        -------------------------------------------------------------- '''
        ii0 = 1 # index of exposure time in keywords
        self.read_keyword(ii0)
        self.fps = self.kwds[ii0]['value']
        return self.fps

    # =====================================================================
    def get_ndr(self,):
        ''' --------------------------------------------------------------
        SCExAO specific: returns the frame rate (from keyword)
        -------------------------------------------------------------- '''
        ii0 = 2 # index of exposure time in keywords
        self.read_keyword(ii0)
        self.ndr = self.kwds[ii0]['value']
        return self.ndr

    # =====================================================================
    def get_crop(self,):
        ''' --------------------------------------------------------------
        SCExAO specific: returns the frame rate (from keyword)
        -------------------------------------------------------------- '''
        ii0 = 3 # index of exposure time in keywords
        self.crop = np.zeros(4)
        for i in range(4):
            self.read_keyword(ii0+i)
            self.crop[i] = self.kwds[ii0+i]['value']
        return self.crop

# =================================================================
# =================================================================



import numpy as np
from scipy.ndimage import rotate

dtor = np.pi / 180.0  # convert degrees to radians


# =========================================================================
def _xyic(ys, xs, between_pix=False):
    ''' --------------------------------------------------------------
    Private utility: returns two arrays of (y, x) image coordinates

    Array values give the pixel coordinates relative to the center of
    the array, that can be offset by 0.5 pixel if between_pix is set
    to True

    Parameters:
    ----------
    - ys: (integer) vertical size of the array (in pixels)
    - xs: (integer) horizontal size of the array (in pixels)
    - between_pix: (boolean) places center of array between pixels

    Returns:
    -------
    A tuple of two arrays: (yy, xx) ... IN THAT ORDER!
    -------------------------------------------------------------- '''
    offset = 0
    if between_pix is True:
        offset = 0.5
    xx = np.outer(np.ones(ys), np.arange(xs)-xs//2+offset)
    yy = np.outer(np.arange(ys)-ys//2+offset, np.ones(xs))
    return (yy, xx)


# =========================================================================
def _dist(ys, xs, between_pix=False):
    ''' --------------------------------------------------------------
    Private utility: returns a distance 2D array

    Array values give the distance to array center, that can be offset
    by 0.5 pixel if between_pix is set to True

    Parameters:
    ----------
    - ys: (integer) vertical size of the array (in pixels)
    - xs: (integer) horizontal size of the array (in pixels)
    - between_pix: (boolean) places center of array between pixels
    -------------------------------------------------------------- '''
    yy, xx = _xyic(ys, xs, between_pix=between_pix)
    return np.hypot(yy, xx)


# ======================================================================
def hole_mask(sz, hcoords, hrad, pscale, between_pix=True):
    ''' -------------------------------------------------------------
    Returns a square "sz x sz" ndarray of a mask of circular holes

    Parameters:
    ----------
    - sz      : size of the 2D array to produce (in pixels)
    - hcoords : array of (x,y) hole coordinates in *unit* (float)
    - hrad    : sub-aperture radius in *unit* (float)
    - pscale  : pixel scale in *unit* per pixel

    Notes:
    -----
    The *unit* choice is up to the user but for most programs,
    using meters (for an actual telescope) sounds sensible.
    ------------------------------------------------------------- '''
    pup = np.zeros((sz, sz))
    hole = uniform_disk(sz, sz, hrad/pscale, between_pix=between_pix)

    xs = np.round(hcoords[:, 0]/pscale).astype(int)
    ys = np.round(hcoords[:, 1]/pscale).astype(int)

    for ii in range(xs.size):
        pup += np.roll(hole, (xs[ii], ys[ii]), axis=(1, 0))

    return pup


# ==================================================================
def get_prad(fsz=512, ldim=8.0, wl=1.6e-6, pscale=10.0):
    ''' ----------------------------------------------------------
    returns the pixel size that, in a Fourier simulation,
    would result in the requested sampling requirements:

    - fsz    : size of the square Fourier array (default: 512x512)
    - ldim   : linear dimension in meters (default: 8.0)
    - wl     : wavelength in meters (default: 1.6e-6)
    - pscale : plate scale in mas / pixel (default: 10.0)
    ---------------------------------------------------------- '''
    ld = wl/ldim * 648e6 / np.pi / pscale  # l/D in pixels
    return(fsz/ld/2)


# ==================================================================
def spectral_sampling(wl1, wl2, nl, wavenum=False):
    ''' ----------------------------------------------------------
    Returns an array of regularly sampled wavelength between
    wl1 and wl2, optionally equally spaced in wave numbers.
    - wl1: shorter wavelength (in meters)
    - wl2: longer wavelength (in meters)
    - nl: number of spectral bins
    - wavenum: True if sampling equally spaced in wave numbers
    ---------------------------------------------------------- '''
    if (not wavenum):
        return (wl1 + (wl2-wl1) * np.arange(nl) / (nl-1))
    else:
        k1, k2 = 1./wl1, 1./wl2
        kk = k1 + (k2-k1) * np.arange(nl) / (nl-1)
        return (1./kk)


# ==================================================================
def ring_grid_coords(nel=6, rad=10.0, central=True, rot=0.0):
    ''' ----------------------------------------------------------
    returns a 2D array of "nel" real set of x,y coordinates for
    points located on a circle of radius "rad". Optionally, the
    original central point can be discarded and the whole thing
    can be rotated by an angle "rot" in radians.

    Parameters:
    ----------
    - nel     : the number of points along the circle
    - rad     : the radius of the ring (float)
    - central : include or not the central point (boolean)
    - rot     : a rotation angle (in radians)
    ---------------------------------------------------------- '''
    th = 2*np.pi * np.arange(nel)/float(nel) + rot
    ring = np.array([rad * np.cos(th), rad * np.sin(th)])
    if central is False:
        return ring
    else:
        oring = np.zeros((2, nel+1))
        oring[:, 1:] = ring
        return oring


# ==================================================================
def meta_ring_grid_coords(xy, nel=6, rad=10.0, central=True, rot=0.0):
    '''---------------------------------------------------------------
    returns a single 2D array of real x,y coordinates for a ring of
    nel point surrounding each point (x,y) coordinate provided in the
    input xy array.

    Parameters;
    ----------
    - xy      : a 2D array of (x,y) coordinates (float)
    - nel     : the number of points along the circle
    - rad     : the radius of the ring (float)
    - central : include or not the central point (boolean)
    - rot     : a rotation angle of circle coords (in radians)
    -------------------------------------------------------------- '''

    xs = np.array(())
    ys = np.array(())

    npt = np.max(xy.shape)  # number of points in the input grid
    xy0 = xy.copy()  # local copy
    if xy0.shape[0] == npt:
        xy0 = xy0.T

    temp = ring_grid_coords(nel=nel, rad=rad, central=central, rot=rot)

    for k in range(npt):
        xs = np.append(xs, temp[0, :] + xy0[0, k])
        ys = np.append(ys, temp[1, :] + xy0[1, k])

    if xy.shape[0] == npt:
        return(np.array((xs, ys)).T)
    else:
        return(np.array((xs, ys)))


# ==================================================================
def hex_grid_coords(nr=1, radius=10, rot=0.0):
    ''' ----------------------------------------------------------
    returns a 2D array of real x,y coordinates for a regular
    hexagonal grid that fits within a hexagon.

    Parameters:
    ----------
    - nr     : the number of "rings" (integer)
    - radius : the radius of a ring (float)
    - rot    : a rotation angle (in radians)
    ---------------------------------------------------------- '''
    rotd = rot * np.pi / 180
    RR = np.array([[np.cos(rotd), -np.sin(rotd)],
                   [np.sin(rotd),  np.cos(rotd)]])

    ij0 = np.linspace(-nr, nr, 2*nr+1)
    ii, jj = np.meshgrid(ij0, ij0)
    xx = radius * (ii + 0.5 * jj)
    yy = radius * jj * np.sqrt(3)/2
    cond = np.abs(ii + jj) <= nr
    return RR.dot(np.array((xx[cond], yy[cond])))


# =============================================================================
def elt_grid_coords(rr=1.45, rot=0.0):
    ''' -----------------------------------------------------------
    returns the coordinates of active segments of the ELT

    Parameters:
    ----------
    - rr: the pitch of the segments (float)

    Note:
    ----
    - for the actual ELT, the picth is equal to 1.45 meters
    ----------------------------------------------------------- '''
    nr = 18  # for the ELT, no choice!
    no = 4  # idem for the central obstruction
    xx, yy = hex_grid_coords(nr=nr, radius=rr, rot=rot)
    xxo, yyo = hex_grid_coords(nr=no, radius=rr, rot=rot)

    for ii, test in enumerate(xxo):
        throw = (xxo[ii] == xx) * (yyo[ii] == yy)
        xx = np.delete(xx, throw)
        yy = np.delete(yy, throw)

    keep = np.sqrt(xx**2+yy**2) < (nr - 0.1) * rr * np.sqrt(3) / 2

    return xx[keep], yy[keep]


# ==================================================================
def meta_hex_grid_coords(xy, nr=1, radius=10):
    '''---------------------------------------------------------------
    returns a single 2D array of real x,y coordinates for a regular
    hexagonal grid of nr rings around each point (x,y) coordinate
    provided in the input xy array.

    Intended use: build a hex grid of hex grids, for instance to help
    with the discretization of a hexagonal segmented aperture.

    Parameters;
    ----------
    - xy    : a 2D array of (x,y) coordinates (float)
    - nr    : the number of hex "rings" to be created around each point
    - radius: the radius of a ring (float)
    ------------------------------------------------------------------- '''

    xs = np.array(())
    ys = np.array(())

    npt = np.max(xy.shape)  # number of points in the input grid
    xy0 = xy.copy()  # local copy
    if xy0.shape[0] == npt:
        xy0 = xy0.T

    temp = hex_grid_coords(nr, radius)
    for k in range(npt):
        xs = np.append(xs, temp[0, :] + xy0[0, k])
        ys = np.append(ys, temp[1, :] + xy0[1, k])

    if xy.shape[0] == npt:
        return(np.array((xs, ys)).T)
    else:
        return(np.array((xs, ys)))


# ==================================================================
def hex_mirror_model(nra, nrs, step, fill=False, rot=0.0, cobs=True):
    ''' -------------------------------------------------------------------
    - nra : the number of rings for the global aperture (rings)
    - nrs : the number of rings per segment
    - step: the minimum distance between two segments centers (float)
    - fill: adds points at gaps between segments and on the edges
    - cobs: central obstruction (central segment missing?) (bool)
    ------------------------------------------------------------------- '''

    RR = np.array([[np.cos(rot), -np.sin(rot)],
                   [np.sin(rot),  np.cos(rot)]])

    # the (coarse) *array* geometry
    # =============================
    seg1 = hex_grid_coords(nra, step)
    if cobs is True:
        keep = (np.abs(seg1[0, :]) > 1e-3) + (np.abs(seg1[1, :]) > 1e-3)
        seg1 = seg1[:, keep]

    radius = step / (np.sqrt(3) * nrs)
    if fill is True:
        nrs += 1
    res = RR.dot(meta_hex_grid_coords(np.flipud(seg1), nrs, radius))
    return(res)


# ==================================================================
def F_test_figure(ys, xs, ww):
    ''' ------------------------------------------------------------
    Returns an (ys x xs) size array with an uppercase F drawn inside
    The F letter overall size is 5x3 times the thickness of the line.

    To be used when lost in the woods, when it comes to orienting
    arrays and images, for instance after Fourier Transforms.

    The orientation convention: origin in the bottom left corner, to
    get the F to be upright.

    Parameters:
    ----------
    - (ys, xs): the dimensions of the array
    - ww      : the thickness of the line
    --------------------------------------------------------------------
    '''
    res = np.zeros((ys, xs))
    xx, yy = np.meshgrid(np.arange(xs)-xs/2+ww/2, np.arange(ys) - ys/2 + ww/2)
    res[(xx > -ww) * (xx <= 0) * (yy > -2*ww) * (yy <= 3*ww)] = 1.0
    res[(xx >= 0) * (xx <= 2*ww) * (yy > 2*ww) * (yy <= 3*ww)] = 1.0
    res[(xx >= 0) * (xx <= ww) * (yy > 0) * (yy <= ww)] = 1.0

    return(res)


# ==================================================================
def uniform_rect(ys, xs, yw, xw, between_pix=True):
    ''' ---------------------------------------------------------
    returns an (ys x xs) array with a centered uniform rectangle
    of size (yw x xw)

    Parameters:
    ----------
    - ys, xs: the size of the array
    - yw, xw: the size of the rectangle
    - between_pix: flag
    ---------------------------------------------------------  '''
    yy, xx = _xyic(ys, xs, between_pix=between_pix)
    res = np.zeros_like(xx)
    res[(np.abs(xx) <= xw/2)*(np.abs(yy) <= yw/2)] = 1.0

    return(res)


# =========================================================================
def uniform_disk(ys, xs, radius, between_pix=False):
    ''' Returns a centered 2D uniform disk array
    ----------------------------------------------
    Parameters:
    - (ys, xs)   : array size
    - radius     : radius of the disk
    - between_pix: (boolean) center between pixels
    ---------------------------------------------- '''
    mydist = _dist(ys, xs, between_pix=between_pix)
    ud = np.zeros_like(mydist)
    ud[mydist <= radius] = 1.0
    return ud


# ==================================================================
def uniform_hex(ys, xs, hrad, between_pix=False):
    ''' ---------------------------------------------------------
    returns an (ys x xs) array with a uniform hexagon of radius "hrad"
    ---------------------------------------------------------  '''
    yy, xx = _xyic(ys, xs, between_pix=between_pix)
    res = np.ones_like(xx)
    res[np.abs(yy) > hrad - xx / np.sqrt(3)] = 0.0
    res[np.abs(yy) > hrad + xx / np.sqrt(3)] = 0.0
    res[np.abs(xx) > 0.5 * np.sqrt(3) * hrad] = 0.0
    return res


# ==================================================================
def four_spider_mask(ys, xs, pix_rad, pdiam, odiam=0.0,
                     beta=45.0, thick=0.25, offset=0.0,
                     spiders=True, split=False, between_pix=True):
    ''' ---------------------------------------------------------
    tool function called by other routines to generate specific
    pupil geometries. Although the result is scaled by pix_rad in
    pixels, telescope specifics are provided in meters.

    Parameters:
    ----------
    - ys, xs  : dimensions of the 2D array      (in pixels)
    - pix_rad : radius of the circular aperture (in pixels)
    - pdiam   : diameter of the aperture        (in meters)
    - odiam   : diameter of the obstruction     (in meters)
    - beta    : angle of the spiders            (in degrees)
    - thick   : thickness of the spiders        (in meters)
    - offset  : spider intersect point distance (in meters)
    - spiders : flag to true to include spiders (boolean)
    - split   : split the mask into four parts  (boolean)
    --------------------------------------------------------- '''

    beta = beta * dtor  # converted to radians
    ro = odiam / pdiam
    yy, xx = _xyic(ys, xs, between_pix=between_pix)
    mydist = np.hypot(yy, xx)

    thick *= pix_rad / pdiam
    offset *= pix_rad / pdiam

    x0 = thick/(2 * np.sin(beta)) + offset
    y0 = thick/(2 * np.cos(beta)) - offset * np.tan(beta)

    if spiders:
        # quadrants left - right
        a = ((xx >= x0) * (np.abs(np.arctan(yy/(xx-x0+1e-8))) < beta))
        b = ((xx <= -x0) * (np.abs(np.arctan(yy/(xx+x0+1e-8))) < beta))
        # quadrants up - down
        c = ((yy >= 0.0) * (np.abs(np.arctan((yy-y0)/(xx+1e-8))) > beta))
        d = ((yy < 0.0) * (np.abs(np.arctan((yy+y0)/(xx+1e-8))) > beta))

    # pupil outer and inner edge
    e = (mydist <= np.round(pix_rad))
    if odiam > 1e-3:  # threshold at 1 mm
        e *= (mydist > np.round(ro * pix_rad))

    if split:
        res = np.array([a*e, b*e, c*e, d*e])
        return(res)

    if spiders:
        return((a+b+c+d)*e)
    else:
        return(e)


# ======================================================================
def four_quadrant_model_split(mask_xy, beta=51.75, thick=0.45, offset=1.28):
    ''' -------------------------------------------------------------
    Tool that corresponds takes a set of coordinates (model mask)
    such as the ones used by XARA, and given the spider geometry
    parameters, splits it into quadrants.

    Parameters:
    ----------

    - mask_xy: a (Nx2) array of (x,y) coordinate points (in meters)
    - beta: spider angle (in degrees)
    - thick: the thickness of the spiders (in meters)
    - offset: spider intersect offset (in meters)

    Note:
    ----

    Default values are set for SCExAO.
    ------------------------------------------------------------- '''
    xx = mask_xy[:, 0]
    yy = mask_xy[:, 1]

    beta1 = beta * dtor

    x0 = thick / (4 * np.sin(beta1)) + 0.5 * offset
    y0 = thick / (4 * np.cos(beta1)) - 0.5 * offset * np.tan(beta1)

    quad1 = (xx >= 0) * (np.abs(np.arctan(yy/(xx-x0+1e-8))) < beta1)
    quad2 = (xx <= -x0) * (np.abs(np.arctan(yy/(xx+x0+1e-8))) < beta1)
    quad3 = (yy >= 0.0) * (np.abs(np.arctan((yy-y0)/(xx+1e-8))) > beta1)
    quad4 = (yy <= 0.0) * (np.abs(np.arctan((yy+y0)/(xx+1e-8))) > beta1)

    res = np.array([[xx[quad1], yy[quad1]],
                    [xx[quad2], yy[quad2]],
                    [xx[quad3], yy[quad3]],
                    [xx[quad4], yy[quad4]]])
    return(res)


# ======================================================================
def lwe_mode_vector(split_xyq, iQuad, iMode):
    ''' -------------------------------------------------------------
    Tool that buids a vector containing a LWE mode of index i0, for
    a given split pupil coordinates model, resulting from the
    four_quadrant_model_split() function.

    Parameters:
    ----------

    - split_xyq : array of xy - coordinates (in meters)
    - iQuad     : index of quadrant (0 - 4)
    - iMode     : index of mode (0: piston, 1: tip, 2: tilt)
    ------------------------------------------------------------- '''

    nq = split_xyq.shape[0]

    vector = np.array([])
    for iq in range(nq):
        if iq != iQuad:
            vector = np.append(vector, np.zeros(split_xyq[iq, 0].size))
        else:
            if iMode == 0:
                vector = np.append(vector, np.ones(split_xyq[iq, 0].size))
            elif iMode == 1:
                vector = np.append(vector, split_xyq[iq, 0])
            else:
                vector = np.append(vector, split_xyq[iq, 1])

    vector /= vector.std()
    return(vector)


# ======================================================================
def HST(xs, ys, radius, spiders=True, between_pix=True):
    ''' -------------------------------------------------------------
    Draws the Hubble Space Telescope pupil of given radius in a array

    Parameters:
    ----------

    - (xs, ys) : dimensions of the 2D array
    - radius   : outer radius of the aperture (in pixels)
    - spiders  : boolean (w/ or w/out spiders)
    -------------------------------------------------------------  '''
    # pupil description
    pdiam, odiam = 2.4, 0.792  # tel. and obst. diameters (meters)
    thick = 0.20               # adopted spider thickness (meters)
    beta = 45.0                # spider angle
    offset = 0.0
    return(four_spider_mask(ys, xs, radius, pdiam, odiam,
                            beta=beta, thick=thick, offset=offset,
                            spiders=spiders, between_pix=between_pix))


# ======================================================================
def HST_NIC1(PSZ, rad, between_pix=True, ang=0):
    ''' ---------------------------------------------------------
    returns an array that draws the pupil of HST/NICMOS1 camera

    Parameters:
    ----------
    - PSZ:     size of the array (assumed to be square)
    - rad:     radius of the standard pupil (in pixels)
    - between_pix: flag
    - ang:     global rotation of the pupil (in degrees)

    Remarks:
    -------
    This fairly complex procedure attempts to reproduce the
    features of the telescope and that of the cold mask inside
    NICMOS, including the misalignment between the two.

    This was used in the following publication:
    https://ui.adsabs.harvard.edu/abs/2020A&A...636A..72M/abstract
    --------------------------------------------------------- '''
    yy, xx = _xyic(PSZ, PSZ, between_pix=between_pix)
    mydist = np.hypot(yy, xx)

    NCM = np.zeros_like(mydist)  # nicmos cold mask

    # --------------------------------
    # OTA: Optical Telescope Assembly
    # --------------------------------
    OTA = np.zeros_like(mydist)
    OTA[mydist <= 1.000 * rad] = 1.0     # outer radius
    OTA[mydist <= 0.330 * rad] = 0.0     # telescope obstruction
    OTA[np.abs(xx) < 0.011 * rad] = 0.0  # spiders
    OTA[np.abs(yy) < 0.011 * rad] = 0.0  # spiders

    tmp = np.roll(mydist, int(0.8921 * rad), axis=1)
    OTA[tmp <= 0.065 * rad] = 0.0        # mirror pad

    tmp = np.roll(
        np.roll(mydist, int(0.7555 * rad), axis=0),
        int(-0.4615 * rad), axis=1)
    OTA[tmp <= 0.065 * rad] = 0.0        # mirror pad

    tmp = np.roll(
        np.roll(mydist, int(-0.7606 * rad), axis=0),
        int(-0.4564 * rad), axis=1)
    OTA[tmp <= 0.065 * rad] = 0.0        # mirror pad

    # --------------------------------
    # NCM: NICMOS COLD MASK
    # --------------------------------
    NCM = np.zeros_like(mydist)           # nicmos cold mask
    NCM[mydist <= 0.955 * rad] = 1.0      # outer radius
    NCM[mydist <= 0.372 * rad] = 0.0      # obstruction 0.372
    NCM[np.abs(xx) < 0.0335 * rad] = 0.0  # fat spiders
    NCM[np.abs(yy) < 0.0335 * rad] = 0.0  # fat spiders

    # PADS
    cpadr = 0.065
    NCM[(xx >= (0.8921-cpadr)*rad) * (np.abs(yy) <= cpadr*rad)] = 0.0
    xx1 = rotate(xx, 121, order=0, reshape=False)
    yy1 = rotate(yy, 121, order=0, reshape=False)
    NCM[(xx1 >= (0.8921-cpadr)*rad) * (np.abs(yy1) <= cpadr*rad)] = 0.0
    xx1 = rotate(xx, -121.5, order=0, reshape=False)
    yy1 = rotate(yy, -121.5, order=0, reshape=False)
    NCM[(xx1 >= (0.8921-cpadr)*rad) * (np.abs(yy1) <= cpadr*rad)] = 0.0

    NCM = np.roll(
        np.roll(NCM, int(-0.0 * rad), axis=1),
        int(-0.08 * rad), axis=0)  # MASK SHIFT !!
    res = 1.0 * (OTA * NCM)

    if ang != 0:
        res = rotate(res, ang, order=0, reshape=False)
    return res


# ==================================================================
def KBENCH(sz, pscale=100.0, noc=False, between_pix=False):
    ''' ---------------------------------------------------------
    Returns a square (sz x sz) array filled with the aperture
    of the KERNEL segmented DM.

    Parameters:
    ----------
    - sz     : size of the array (integer)
    - pscale : pupil pixel scale in micron / pixel (float)
    - noc    : masking central segment (default: False)

    Planned: (but not implemented yet)
    -------
    - nodead : masking dead segments (default: False)
    --------------------------------------------------------- '''
    sstep = 0.650  # segment step size (according to BMC doc)
    yy, xx = _xyic(sz, sz, between_pix=between_pix)
    res = np.zeros_like(xx)

    seg_rad = int(sstep / pscale)  # segment radius in pixels
    scoords = hex_grid_coords(
        nr=8, radius=seg_rad, rot=2*np.pi/3).astype('int')

    if noc:  # mask central segment
        scoords = np.delete(scoords, scoords.shape[1]//2, axis=1)

    seg0 = uniform_hex(sz, sz, seg_rad/np.sqrt(3)-1)
    seg0 = seg0.T

    for ii in range(scoords.shape[1]):
        res += np.roll(np.roll(seg0, scoords[0, ii], axis=0),
                       scoords[1, ii], axis=1)
    return res


# ==================================================================
def JWST(sz, pscale=0.1, aperture="CLEARP",
         parx=0.02579476, pary=0.01289738):
    ''' ---------------------------------------------------------
    Returns a square (sz x sz) array filled with a representation
    of the JWST aperture.

    Note: PAR stands for pupil alignment reference. It is what
    makes the difference between CLEAR and CLEARP

    Parameters:
    ----------
    - sz       : size of the array (integer)
    - pscale   : pupil pixel scale in meter / pixel (float)
    - aperture : identifier (string: default "CLEARP")
    - parx     : hrztal PAR misalignment (default=0.02579476)
    - pary     : vrtcal PAR misalignment (default=0.01289738)

    Remarks:
    -------
    Using the longer wavelength filters ("F277W" and longer)
    requires the insertion of the pupil alignment reference (PAR)
    which alters the pupil shape from CLEAR to CLEARP.

    CLEARP is assumed by default. Anything else will use CLEAR.
    --------------------------------------------------------- '''
    bpix = (sz % 2 == 0)
    yy, xx = _xyic(sz, sz, between_pix=bpix)
    sp_th = 0.125  # true thickness spider in meters
    trad = 1.32   # true segment radius in meters
    orad = 1.02   # true PAR obstruction radius in meters
    SP_th = 0.33  # true PAR spider thickness in meters

    seg_rad = int(trad / pscale)   # segment radius in pixels
    spi_hthk = sp_th / pscale / 2  # TEL spider half-thickness in pixels
    SPI_hthk = SP_th / pscale / 2  # PAR spider half-thickness in pixels
    PAR_rad = orad / pscale        # PAR obstruction radius in pixels

    res = np.zeros((sz, sz))

    # segmented aperture first
    # pop central segment!
    # ------------------------
    scoords = hex_grid_coords(nr=2, radius=seg_rad, rot=0)
    scoords = np.delete(scoords, scoords.shape[1]//2, axis=1)
    scoords = scoords.astype('int')

    seg0 = uniform_hex(sz, sz, seg_rad/np.sqrt(3)-1)
    seg0 = seg0.T

    for ii in range(scoords.shape[1]):
        res += np.roll(np.roll(seg0, scoords[0, ii], axis=0),
                       scoords[1, ii], axis=1)

    # adding spider arm shadow
    # ------------------------
    sp1 = 1 - (np.abs(xx) <= spi_hthk) * (yy > 0)
    sp2 = rotate(sp1, 150.0, order=0, reshape=False)
    sp3 = np.fliplr(sp2)

    res *= sp1 * sp2 * sp3

    # CLEARP or CLEAR?
    # ----------------
    if aperture == "CLEARP":
        mask_obs = 1-uniform_disk(sz, sz, PAR_rad)

        offx = int(np.round(parx/pscale))
        offy = int(np.round(pary/pscale))

        sp1 = 1 - (np.abs(xx) <= SPI_hthk) * (yy > 0)
        sp2 = rotate(sp1, 120.0, order=0, reshape=False)
        sp3 = np.fliplr(sp2)

        mask_obs *= sp1 * sp2 * sp3
        mask_obs = np.roll(mask_obs, (offx, offy), axis=(1, 0))
        res *= mask_obs
    return res


# ==================================================================
def JWST_NRM(sz, pscale=0.0064486953125):
    ''' ---------------------------------------------------------
    Returns a square (sz x sz) array filled with a representation
    of the JWST aperture with the non-redundant mask.

    Parameters:
    ----------
    - sz       : size of the array (integer)
    - pscale   : pupil pixel scale in meter / pixel (float)

    Credit:
    ------
    Written by J. Marquez (2020)
    --------------------------------------------------------- '''

    # x,y hole coordinates (meters)
    X_m = [1.99, 1.32, 0.0, 1.99, 1.32, -1.32, -2.64]
    Y_m = [-1.14583, -2.28895, -2.28895, 1.14583, 2.28895, 2.28895, 0.0]

    hole_radius = 0.4650001189  # radius hole (meter)
    hexa_f = np.zeros([sz, sz])  # Final NRM

    # First Hole
    # ----------
    bpix = (sz % 2 == 0)
    hexa = uniform_hex(sz, sz, hole_radius/pscale, between_pix=bpix).T

    for i in range(len(X_m)):
        hexa_p = np.roll(np.roll(
            hexa, int(X_m[i]/pscale), 0),
                         int(Y_m[i]/pscale), 1)
        hexa_f += hexa_p
    return(hexa_f)


# ==================================================================
def PHARO(PSZ, rad, mask="std", between_pix=True, ang=0):
    ''' ---------------------------------------------------------
    returns an array that draws the pupil of the PHARO camera
    of the Palomar Hale telescope, at the center of an array of
    size "PSZ" with radius "radius" (both in pixels).

    Parameters:
    ----------
    - PSZ:     size of the array (assumed to be square)
    - rad:     radius of the standard pupil (in pixels)
    - mask:    aperture mask used
      + "std": standard cross (default)
      + "med": medium cross
    - between_pix: flag
    - ang:     global rotation of the pupil (in degrees)

    Notes:
    -----
    The reference is the standard cross radius, which corresponds
    to an actual 4.978 m clear aperture on the telescope. The
    central obstruction diameter is then 1.841 m (standard cross).

    The med_cross corresponds to a 4.646 m clear aperture,
    and a 2.293 central obstruction.

    See Hayward et al (2000) in PASP for reference.
    --------------------------------------------------------- '''
    yy, xx = _xyic(PSZ, PSZ, between_pix=between_pix)
    mydist = np.hypot(yy, xx)

    res = np.zeros_like(mydist)

    if "med" in mask.lower():
        res[mydist <= 0.933 * rad] = 1.0    # undersized aperture
        res[mydist <= 0.460 * rad] = 0.0    # oversized obstruction
        res[np.abs(xx) < 0.05 * rad] = 0.0  # fat spiders
        res[np.abs(yy) < 0.05 * rad] = 0.0  # fat spiders

    else:
        res[mydist <= rad] = 1.0   # std clear aperture
        res[mydist <= 0.370 * rad] = 0.0
        res[np.abs(xx) <= 0.015 * rad] = 0.0
        res[np.abs(yy) <= 0.015 * rad] = 0.0

    if ang != 0:
        res = rotate(res, ang, order=0, reshape=False)
    return res


# ==================================================================
def VLT(n, m, radius, spiders=True, between_pix=True):
    ''' ---------------------------------------------------------
    returns an array that draws the pupil of the VLT
    at the center of an array of size (n,m) with radius "radius".

    Parameters describing the pupil were deduced from a pupil mask
    description of the APLC coronograph of SPHERE, by Guerri et al,
    2011.

    http://cdsads.u-strasbg.fr/abs/2011ExA....30...59G
    --------------------------------------------------------- '''

    # VLT pupil description
    # ---------------------
    pdiam, odiam = 8.00, 1.12  # tel. and obst. diameters (meters)
    thick = 0.04               # adopted spider thickness (meters)
    offset = 1.11              # spider intersection offset (meters)
    beta = 50.5                # spider angle beta

    return(four_spider_mask(m, n, radius, pdiam, odiam,
                            beta, thick, offset, spiders=spiders,
                            between_pix=between_pix))


# ==================================================================
def SPHERE_IRDIS_SAM(sz, ldim=8.0):
    ''' ---------------------------------------------------------
    Returns a square (sz x sz) array featuring a 2D "image" of
    the Sparse Aperture Mask (SAM) inside SPHERE's IRDIS camera.

    Parameters:
    ----------
    - sz: pixel size of the array (integer)
    - ldim: linear dimension of the image (in meters)

    Remarks:
    -------
    The geometry of the mask was retroengineered from technical
    specs found in the SPHERE instrument documentation and the
    analysis of actual SPHERE SAM frames. There seems to be a
    discrepancy with the published geometry but I tend to trust
    my analysis.
    --------------------------------------------------------- '''

    pscale = ldim/sz  # pixel scale (in meters/pixel)
    hrad = 0.5  # circular hole radius (in meters)

    # SAM hole x,y coordinates (in meters)
    hcoords = 0.8 * np.array(
        [[+0.000, -3.9366],
         [-3.788, -1.7496],
         [+3.788, -1.7496],
         [-1.894, -0.6561],
         [-3.788,  0.4374],
         [-1.894,  3.7179],
         [+1.894,  3.7179]])

    pmask = hole_mask(sz, hcoords, hrad, pscale, between_pix=False)
    return pmask


# ==================================================================
def ELT(sz, pscale=0.05, spiders=True):
    ''' ------------------------------------------------------------
    Returns a square (sz x sz) array filled with aa representation
    of the ELT full aperture.

    Parameters:
    ----------
    - sz     : size of the array (integer number of pixels)
    - pscale : pupil pixel scale in meter / pixel (float)
    - spider : includes the spiders if True (boolean)
    ------------------------------------------------------------ '''
    spitch = 1.45 * np.sqrt(3)/2  # segment pitch (in meters)
    sz0 = int(31 * spitch / pscale)
    if sz < sz0:
        print(f"Min array size sz = {sz0} for pscale = {pscale}")
        return

    pup = np.zeros((sz, sz), dtype=float)
    ssz = np.round(spitch * 2 / np.sqrt(3) / pscale).astype(int)
    seg = uniform_hex(sz, sz, ssz//2, (sz % 2 == 0))

    xx, yy = elt_grid_coords(rr=spitch/pscale)

    for ii in range(len(xx)):
        pup += np.roll(
            seg,
            (np.round(xx[ii]).astype(int), np.round(yy[ii]).astype(int)),
            axis=(1, 0))

    if spiders:
        spwidth = 0.51  # spider witdh (in meters)
        sz = pup.shape[0]
        yy, xx = _xyic(sz, sz)
        sp1 = np.abs(yy) > (spwidth / pscale / 2)
        sp2 = rotate(sp1, 120, order=0, reshape=False)
        sp3 = np.fliplr(sp2)
        pup *= sp1 * sp2 * sp3
    return pup


# ==================================================================
def keck(sz, spiders=True, between_pix=True):
    tmp = segmented_aperture(sz, 3, sz//7, rot=0.0)
    yy, xx = _xyic(sz, sz, between_pix=between_pix)
    dist = np.hypot(yy, xx)
    obst = np.ones((sz, sz))

    obst[dist <= 0.121 * sz] = 0.0
    tmp *= obst

    thick = 0.00116 * sz
    for ii in range(6):
        ang = 30.0 + 60.0 * ii
        zone = (xx > 0) * (np.abs(yy) <= thick)
        res = np.ones((sz, sz))
        res[zone] = 0.0
        res = rotate(res, ang, order=0, reshape=False)
        tmp *= res
    return(tmp)


# ==================================================================
def subaru(n, m, radius, spiders=True, between_pix=True):
    ''' ---------------------------------------------------------
    returns an array that draws the pupil of the Subaru Telescope
    at the center of an array of size (n,m) with radius "radius".

    Symbols and values used for the description come from the
    document sent to Axsys for the fabrication of the SCExAO
    PIAA lenses, circa 2009.
    --------------------------------------------------------- '''

    # Subaru pupil description
    # ------------------------
    pdiam, odiam = 7.92, 2.3  # tel. and obst. diameters (meters)
    thick = 0.25              # adopted spider thickness (meters)
    offset = 1.278            # spider intersection offset (meters)
    beta = 51.75              # spider angle beta

    return(four_spider_mask(m, n, radius, pdiam, odiam=odiam,
                            beta=beta, thick=thick, offset=offset,
                            spiders=spiders,
                            between_pix=between_pix))


# ==================================================================
def subaru_lstop(sz, osize=0.1, between_pix=True):
    ''' ---------------------------------------------------------
    returns an array that draws an oversized Lyot stop for the
    pupil of the Subaru Telescope at the center of an array of
    size (sz, sz) with radius "radius".

    Parameters:
    ----------
    - sz: the array size
    - osize: the oversize (default = 0.1 m)
    - between_pix: if True, array centered between pixels
    --------------------------------------------------------- '''

    # Subaru pupil description
    # ------------------------
    pdiam, odiam = 7.92, 2.3  # tel. and obst. diameters (meters)
    thick = 0.25              # adopted spider thickness (meters)
    offset = 1.278            # spider intersection offset (meters)
    beta = 51.75              # spider angle beta

    return(four_spider_mask(
        sz, sz, (sz//2)*(1-2*osize/pdiam),
        pdiam-2*osize, odiam=odiam+2*osize,
        beta=beta, thick=thick+2*osize, offset=offset,
        spiders=True, between_pix=between_pix))


# ==================================================================
def subaru_dbl_asym(xs, ys, radius, spiders=True, PA1=0.0, PA2=90.0,
                    thick1=0.15, thick2=0.15):
    ''' -------------------------------------------------------------
    Returns a pupil mask with *two* asymmetric arms for two distinct
    position angles.

    Parameters:
    ----------

    - (xs, ys) : dimensions of the 2D array
    - radius   : outer radius of the aperture
    - spiders  : boolean (w or without spiders)
    - PA1      : position angle of arm #1 (in degrees)
    - PA2      : position angle of arm #2 (in degrees)
    - thick1   : asymm. arm #1 thickness (% of aperture diameter)
    - thick2   : asymm. arm #2 thickness (% of aperture diameter)
    ------------------------------------------------------------- '''
    a = subaru_asym(xs, ys, radius, spiders=spiders, PA=PA1, thick=thick1)
    b = subaru_asym(xs, ys, radius, spiders=spiders, PA=PA2, thick=thick2)
    return(a*b)


# ==================================================================
def radial_arm(xs, ys, radius, PA=0.0, thick=0.15):
    ''' -------------------------------------------------------------
    Produces a pupil mask for an occulting radial arm for a given
    position angle and thickness.

    Parameters:
    - xs, ys : dimensions of the 2D array   (integer # of pixels)
    - radius : outer radius of the aperture (integer # of pixels)
    - PA     : position angle of the arm    (degrees)
    - thick  : fraction of the ap. diameter (float)
    ------------------------------------------------------------- '''
    res = np.ones((ys, xs))
    ang = np.mod(PA, 360.0)
    yy, xx = _xyic(ys, xs)
    zone = (xx > 0) * (np.abs(yy) <= thick*radius)
    res[zone] = 0.0
    res = rotate(res, ang, order=0, reshape=False)
    return(res.astype('int'))


# ==================================================================
def subaru_asym(xs, ys, radius, spiders=True, PA=0.0, thick=0.15):
    ''' -------------------------------------------------------------
    Returns a pupil mask with an asymmetric arm that mostly follows
    the geometry of the original APF-WFS.

    Parameters:
    ----------

    - xs, ys  : dimensions of the 2D array
    - radius  : outer radius of the aperture
    - spiders : boolean (w or without spiders)
    - PA      : position angle of the arm (in degrees)
    - thick   : asymm. arm thickness (% of aperture diameter)
    ------------------------------------------------------------- '''

    pup = subaru(xs, ys, radius, spiders)
    arm = radial_arm(xs, ys, radius, PA=PA, thick=thick)

    return(pup * arm)


# ======================================================================
def segmented_aperture(sz, nr, srad, rot=0.0):
    ''' ----------------------------------------------------------------
    Returns the square sz x sz image of a segmented aperture.

    Parameters:
    ----------
    - sz   : the size of the array
    - nr   : the number of rings (annuli) making up the aperture
    - srad : the radius of a segment (in pixels)
    - rot  : a rotation angle (in radians)
    ---------------------------------------------------------------- '''
    xy = hex_mirror_model(nr+1, srad, srad, rot=rot).astype(np.int)+sz//2
    pup = np.zeros((sz, sz))
    for i in range(xy.shape[1]):
        pup[xy[0, i], xy[1, i]] = 1.0
    return(pup)


# ======================================================================
def golay9_coords(prad, rot=0.0):
    ''' -------------------------------------------------------------
    Returns the holes coordinates of a 9-hole Golay array

    Parameters:
    ----------
    - prad: the radius of the circular aperture to be masked
    - rot: mask azimuth (in radians)
    ------------------------------------------------------------- '''
    dstep = prad / 3.5  # why exactly 3.5 here?
    d1 = dstep*np.sqrt(7)
    th0 = np.arctan(np.sqrt(3)/2)

    xs = np.array([])
    ys = np.array([])

    for i in range(3):
        theta = 2.0 * i * np.pi / 3.0 + rot

        for k in range(2, 4):
            xs = np.append(xs, k*dstep*np.cos(theta))
            ys = np.append(ys, k*dstep*np.sin(theta))

        xs = np.append(xs, d1*np.cos(theta-th0))
        ys = np.append(ys, d1*np.sin(theta-th0))

    return np.array([xs, ys]).T


# ======================================================================
def golay9(sz, prad, hrad, between_pix=True, rot=0.0):
    ''' -------------------------------------------------------------
    Returns a square "sz x sz" NR Golay 9 pupil model

    Parameters:
    ----------
    - sz:          size of the 2D array to produce (in pixels)
    - prad:        size of the pupil the mask fits into (in pixels)
    - hrad:        sub-aperture radius (in pixels)
    - between_pix: centers the array between 4 pixels (boolean)
    - rot:         mask azimuth (in radians)
    ------------------------------------------------------------- '''
    pup = np.zeros((sz, sz))
    hole = uniform_disk(sz, sz, hrad, between_pix=between_pix)
    coords = golay9_coords(prad, rot)

    xs = np.cast['int'](np.round(coords[:, 0]))
    ys = np.cast['int'](np.round(coords[:, 1]))

    for ii in range(xs.size):
        pup += np.roll(np.roll(hole, ys[ii], axis=0), xs[ii], axis=1)

    return(pup)


# ======================================================================
def lwe_mode_bank_2D(sz, odiam=8.0, beta=51.75, offset=1.28):
    ''' -------------------------------------------------------------
    Builds a 3D array containing pupil images of the 12 raw LWE
    modes: piston, tip and tilt, for the four expected quadrants
    of a pupil like Subaru, VLT, ...

    Parameters:
    ----------

    - sz: size of the 2D pupil images to be produced
    - odiam: pupil outer diameter     (default 8.0,   in meters)
    - beta: spider angle              (default 51.75, in degrees)
    - offset: spider intersect offset (default: 1.28, in meters)
    ------------------------------------------------------------- '''
    quads = four_spider_mask(sz, sz, sz/2, odiam, 0.0,
                             beta=beta, thick=0.0, offset=offset,
                             spiders=True, split=True)
    xx, yy = np.meshgrid(np.arange(sz)-sz/2, np.arange(sz)-sz/2)

    nm = 12
    bank = np.zeros((nm, sz, sz))
    for ii in range(nm):
        if ((ii % 3) == 0):
            bank[ii] = 1.0 * quads[ii // 3]
        elif ((ii % 3) == 1):
            temp = xx - xx[quads[ii // 3]].mean()
            bank[ii] = temp * quads[ii // 3]
        elif ((ii % 3) == 2):
            temp = yy - yy[quads[ii // 3]].mean()
            bank[ii] = temp * quads[ii // 3]
        bank[ii] /= bank[ii].std()
    return(bank)


# ======================================================================
def fqpm(sz):
    ''' -------------------------------------------------------------
    Four quadrant phase mask algorithm

    The oddball in this module dedicated to the pupil and not image,
    but I don't want to set up a module just for a bloody function.
    ------------------------------------------------------------- '''
    res = np.ones((sz, sz))
    res[:sz//2, :sz//2] = -1
    res[sz//2:, sz//2:] = -1
    return res



#!/usr/bin/env python3

import functools

from weakref import WeakValueDictionary

import numpy as np
import time
from . import pupil

from .camera import Cam, SHCam, CoroCam
from .DM import DM, HexDM
from .atmosphere import Phscreen
from .pupil import uniform_disk as ud

import matplotlib.pyplot as plt
plt.ion()
plt.show()
plt.rcParams['image.origin'] = "lower"

# ===========================================================
# ===========================================================


class Singleton(type):
    _instances = WeakValueDictionary()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super(Singleton, cls).__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


def singleton(cls):
    instance = [None]

    @functools.wraps(cls)
    def wrapper(*args, **kwargs):
        print("instance[0] = ", instance[0])
        if instance[0] is None:
            instance[0] = cls(*args, **kwargs)
        else:
            print("already instantiated!")
        return instance[0]
    return wrapper


# ===========================================================
# ===========================================================


class Telescope(object):
    ''' Telecope class
    ---------------------------------------------------------------------------
    This may be overengineering at this point but there may be something to
    gained from gathering information that is relevant to a telescope into a
    separate object.

    One interest is that several cameras fed by the same telescope can have
    their own aperture mask, that can add on top of that of the telescope?

    Information stored here:
    Telescope diameter, pupil geometry, and orientation, ...

    And since I am in the process of rewriting this code, why not?
    ---------------------------------------------------------------------------

    '''
    # ==================================================
    def __init__(self, name="", size=256, radius=128, rebin=5):
        ''' Instantiation of a Telescope:

        Parameters:
        ----------
        - name   : a string descriving the telescope
        - size   : size of the array describing the telescope pupil (pixels)
        - radius : pupil radius (in pixels) within the array
        - rebin  : ovsersampling parameter for finer pupil model
        ------------------------------------------------------------------- '''
        self.tname = "EMPTY"    # telescope name
        self.iname = name       # instrument name
        self.pdiam = 1.0        # telescope diameter (meters)
        self.PA = 0.0           # pupil angle (degrees)
        self.rebin = rebin      # rebin for "grey model"
        self.size = size
        self.radius = radius
        self.update_pupil()     # update pupil array

    # ==================================================
    def __str__(self):
        msg = f"Telescope : {self.tname} - diameter : {self.pdiam:.1f} meters\n"
        msg += f"Instrument : - {self.iname}\n"
        return msg

    # ==================================================
    def get_pupil(self,):
        return self.pupil

    # ==================================================
    def update_pupil(self,):
        ''' Update telescope properties based on provided name
        -----------------------------------------------------------------------

        Default config: an unobstructed  1-meter diameter telescope
        -----------------------------------------------------------------------
        '''
        rsz = self.size * self.rebin
        rrad = self.radius * self.rebin

        if "scexao" in self.iname.lower():
            self.tname = "Subaru"
            self.iname = "SCExAO"
            self.pdiam = 7.92  # telescope pupil diameter
            pup = 1.0 * pupil.subaru(
                rsz, rsz, rrad, spiders=True, between_pix=True)

        elif "elt" in self.iname.lower():
            self.tname = "ELT"
            self.iname = "petalometer"
            self.pdiam = 39  # telescope "diameter"
            pup = pupil.ELT(rsz, pscale=self.pdiam/rsz, spiders=True)

        elif "hst" in self.iname.lower():
            self.tname = "HST"
            self.iname = "NICMOS1"
            self.pdiam = 2.4
            self.PA = 45.0
            tmp = pupil.HST_NIC1(2*rsz, rrad, ang=self.PA)
            pup = tmp[rsz//2:rsz//2+rsz, rsz//2:rsz//2+rsz]

        elif "aoc" in self.iname.lower():
            self.tname = "C2PU"
            self.iname = "AOC"
            self.pdiam = 1.0
            pup = pupil.uniform_disk(rsz, rsz, rrad)

        elif "baldr" in self.iname.lower():
            self.tname = "VLTI"
            if "ut" in self.iname.lower():
                self.pdiam = 8.0
                pup = pupil.VLT(rsz, rsz, rrad)
                self.iname = "BALDR - UT"
            else:
                self.pdiam = 1.8
                pup = pupil.VLT(rsz, rsz, rrad)
                self.iname = "BALDR - AT"

        elif "gravity" in self.iname.lower():
            self.tname = "VLT"
            self.iname = "GRAVITY+"
            self.pdiam = 8.0
            self.PA = 0.0
            pup = pupil.VLT(rsz, rsz, rrad)

        elif "pharo" in self.iname.lower():
            self.tname = "Hale"
            self.iname = "PHARO"
            self.pdiam = 4.978  # PHARO standard cross diameter
            self.PA = 0.0
            pup = pupil.PHARO(
                rsz, rrad, mask="std", between_pix=True, ang=0)
            if "med" in self.iname.lower():
                pup = pupil.PHARO(
                    rsz, rrad, mask="med", between_pix=True, ang=-2)

        elif "nirc2" in self.iname.lower():
            self.tname = "Keck2"
            self.iname = "NIRC2"
            self.pdiam = 10.2
            self.PA = -20.5
            th0 = self.PA*np.pi/180.0  # pupil angle
            pup = pupil.segmented_aperture(rsz, 3, int(rrad/3), rot=th0)

        elif "jwst" in self.iname.lower():
            self.tname = "JWST"
            self.pdiam = 6.6
            self.PA = 0.0
            if "nrm" in self.iname.lower():
                pup = pupil.JWST_NRM(rsz, pscale=self.pdiam/rsz)
            else:
                pup = pupil.JWST(rsz, pscale=self.pdiam/rsz, aperture="CLEARP")
            self.iname = "NIRISS"

        elif "kernel" in self.iname.lower():
            self.tname = "BENCH"
            self.iname = "KERNEL"
            self.pdiam = 9.75  # BMC DM largest dimension (in mm)
            self.PA = 0.0
            pup = pupil.KBENCH(rsz, pscale=self.pdiam/rsz)

        else:
            print("Default: unbstructed circular aperture")
            pup = pupil.uniform_disk(rsz, rsz, rrad)

        self.pupil = pup.reshape(
            self.size, self.rebin, self.size, self.rebin).mean(3).mean(1)

# =============================================================================
# =============================================================================


class instrument(metaclass=Singleton):

    # ==================================================
    def __init__(self, name="SCExAO", shdir='/dev/shm/', csz=320):
        print("calling constructor")
        self.name = name
        self.shdir = shdir
        self.delay = 0.1  # sets the default simulation frame rate
        self.csz = csz    # computation size for wavefronts

        self.tel = Telescope(
            name=self.name, size=self.csz, radius=self.csz//2, rebin=5)

        self.cam = None
        self.cam2 = None
        self.DM = None
        self.atmo = None

        # ---------------------------------------------------------------------
        # SCExAO template: DM, atmo and IR camera
        # ---------------------------------------------------------------------
        if "scexao" in self.name.lower():
            print("Creating %s" % (self.name,))
            if "coro" in self.name.lower():
                lstop = pupil.subaru_lstop(self.csz)
                self.add_corono_camera(
                    name="SCExAO_coro", ysz=256, xsz=320, lstop=lstop,
                    pscale=16.7, wl=1.6e-6, slot=1)
            else:
                self.add_imaging_camera(
                    name="SCExAO_chuck", ysz=256, xsz=320,
                    pscale=16.7, wl=1.6e-6, slot=1)

            self.add_membrane_DM(
                dms=50, nch=8, na0=49, iftype="cosine",
                ifr0=1.0, dtype=np.float32)

            self.add_phscreen(
                name="MaunaKea", r0=0.5, L0=10.0, fc=24.5, correc=10.0)

        # ---------------------------------------------------------------------
        # BALDR template: DM, atmosphere & pupil viewing camera
        # ---------------------------------------------------------------------
        elif "baldr" in self.name.lower():
            print(f"Creating {self.name}")
            self.add_corono_camera(name="BALDR", ysz=12, xsz=12, pview=12,
                                   pscale=20, wl=1.6e-6, slot=1)
            # Zernike phase mask
            self.cam.fpm = 1 - (1 - 1j) * ud(csz, csz, self.cam.ld0,
                                             between_pix=True)

            self.add_membrane_DM(
                dms=12, nch=8, na0=10, iftype="cosine", ifr0=1.0)

            self.add_phscreen(
                name="Paranal", r0=0.5, L0=10.0, fc=24.5, correc=1.0)

        # ---------------------------------------------------------------------
        # AOC template: DM, atmosphere and 2 cameras (vis SH + imager IR)
        # ---------------------------------------------------------------------
        elif "aoc" in self.name.lower():
            print("Creating %s" % (self.name,))
            self.add_SH_camera(
                name="AOC_ASO", dsz=128, mls=10, wl=0.8e-6, slot=1)

            self.add_membrane_DM(
                dms=11, nch=8, na0=10, iftype="cone", ifr0=1.0)

            self.add_imaging_camera(
                name="HiPiC", ysz=256, xsz=256, pscale=100, wl=1.6e-6, slot=2)

            self.add_phscreen(
                name="Calern", r0=0.2, L0=10.0, fc=5, correc=1.0)

        # ---------------------------------------------------------------------
        # GRAVITY+ template: DM, atmosphere and 2 cameras (vis SH + imager IR)
        # ---------------------------------------------------------------------
        elif "gravity" in self.name.lower():
            print("Creating %s" % (self.name,))
            self.add_SH_camera(
                name="GRAVITY_WFS", dsz=240, mls=40, wl=0.7e-6, slot=1)

            self.add_membrane_DM(
                dms=40, nch=8, na0=39, iftype="cone", ifr0=1.0)

            self.add_imaging_camera(
                name="VLT_IR", ysz=320, xsz=320, pscale=10.0, wl=1.6e-6,
                slot=2)

            self.add_phscreen(
                name="Paranal", r0=0.5, L0=10.0, fc=20, correc=10.0)

        # ---------------------------------------------------------------------
        # HST NICMOS 1 template: one camera (no atmo, no DM) !
        # ---------------------------------------------------------------------
        elif "hst" in self.name.lower():
            print("Creating %s" % (self.name,))
            self.add_imaging_camera(
                name="HST_NIC1", ysz=84, xsz=84, pscale=43.0, wl=1.9e-6,
                slot=1)

        # ---------------------------------------------------------------------
        # JWST NIRISS template: one camera for now (no DM control yet?)
        # ---------------------------------------------------------------------
        elif "jwst" in self.name.lower():
            print("Creating %s" % (self.name,))
            iname = "NIRISS"
            if "nrm" in self.name.lower():
                iname += "_NRM"
            self.add_imaging_camera(
                name=iname, ysz=80, xsz=80, pscale=65.6, wl=4.8e-6,
                slot=1)

        # ---------------------------------------------------------------------
        # KERNEL bench template: HexDM, and IR camera
        # ---------------------------------------------------------------------
        elif "kernel" in self.name.lower():
            print("Creating %s" % (self.name,))
            self.add_imaging_camera(
                name="kbench", ysz=256, xsz=320, pscale=10.0, wl=1.6e-6,
                slot=1)

            self.add_hex_DM(nr=7, nch=8, na0=15, srad=323.75)

            self.add_phscreen(
                name="Valrose", r0=1.0, L0=10.0, fc=7.5, correc=10.0)

        # ---------------------------------------------------------------------
        # ELT petalometer template:
        # ---------------------------------------------------------------------
        elif "elt" in self.name.lower():
            print("Creating %s" % (self.name,))
            self.add_imaging_camera(
                name="petalometer2.1", ysz=256, xsz=256, pscale=5.0, wl=2.1e-6,
                slot=1)

            self.add_imaging_camera(
                name="petalometer2.3", ysz=256, xsz=256, pscale=5.0, wl=2.3e-6,
                slot=2)

            self.add_membrane_DM(dms=100, nch=4, na0=99,
                                 iftype="cosine", ifr0=1.0)
            self.add_phscreen(
                name="Armazones", r0=0.144, L0=25.0, fc=15, wl=0.5e-6, correc=1.0)

        # ---------------------------------------------------------------------
        # NO template? Go manual.
        # ---------------------------------------------------------------------
        else:
            print("""No template available for '%s':
            assuming that you want a custom configuration...""" % (self.name))

    # ==================================================
    def __str__(self):
        msg = "XAOSIM virtual instrument:\n" + 26*"-" + "\n"
        msg += self.tel.__str__()
        msg += f"\n"

        if self.cam is not None:
            msg += self.cam.__str__()
            return msg

    # ==================================================
    def _install_camera(self, cam, slot=1):
        ''' -------------------------------------------------------------------
        Attempts to integrate a camera to the instrument.

        Checks for the availability of the requested slot.
        Parameters:
        ----------
        - cam: an instance of camera
        - slot: marks the "position" of the camera (1 or 2, default = 1)

        Notes:
        -----
        At the moment, I only anticipate a user to want to have two distinct
        cameras as part of one instrument and only have two slots with hard
        coded names. I could change that and offer the possibility of having
        an ever extensible list of cameras?
        ------------------------------------------------------------------- '''
        if slot == 1:
            if self.cam is None:
                self.cam = cam
                print("Camera installed in slot 1. Handle is xx.cam")
            else:
                print("Camera already in place in slot 1")
        else:
            if self.cam2 is None:
                self.cam2 = cam
                print("Camera installed in slot 2. Handle is xx.cam2")
            else:
                print("Camera already in place in slot 2")

    # ==================================================
    def _get_handle(self, slot=1):
        ''' -------------------------------------------------------------------
        Notes: clearly, it would be better to have a list of slots rather
        than two hardcoded handles... but can I afford this change without
        making some of my current users really angry at me?
        ------------------------------------------------------------------- '''
        if slot == 1:
            try:
                _ = self.cam
                return self.cam
            except AttributeError:
                return None
        else:
            try:
                _ = self.cam2
                return self.cam2
            except AttributeError:
                return None

    # ==================================================
    def _delete_handle(self, slot=1):
        ''' -------------------------------------------------------------------
        Notes: clearly, it would be better to have a list of slots rather
        than two hardcoded handles... but can I afford this change without
        making some of my current users really angry at me?
        ------------------------------------------------------------------- '''
        if slot == 1:
            try:
                del self.cam
                self.cam = None
                print("Handle for slot #%d now free" % (slot,))
            except AttributeError:
                print("Handle for slot #%d already free" % (slot,))
                pass
        else:
            try:
                del self.cam2
                self.cam2 = None
                print("Handle for slot #%d now free" % (slot,))
            except AttributeError:
                print("Handle for slot #%d already free" % (slot,))
                pass

    # ==================================================
    def clear_camera_slot(self, slot=1):
        ''' -------------------------------------------------------------------
        Removes a camera from the instrument.

        Removal is required prior to installing a new camera on a given slot.
        ------------------------------------------------------------------- '''
        tmp = self._get_handle(slot=slot)

        try:
            tmp.stop()
            _ = tmp.shm_cam.fd
            if _ != 0:
                tmp.shm_cam.close()
                self._delete_handle(slot=slot)
        except AttributeError:
            print("Camera slot #%d already clear" % (slot,))
            pass
        del tmp

    # ==================================================
    def add_imaging_camera(self, name="ircam", ysz=256, xsz=320,
                           pscale=16.7, wl=1.6e-6, slot=1):
        ''' -------------------------------------------------------------------
        Adds a conventional imaging camera to the instrument.

        Parameters:
        ----------
        - name   : string describing the camera (used for shm file name)
        - ysz    : vertical size of the detector in pixels (default = 256)
        - xsz    : horizontal size of the detector in pixels (default = 320)
        - pscale : detector plate scale in mas/pixels (default = 16.7)
        - wl     : wavelength of observation in meters (default = 1.6e-6)
        - slot   : marks the "position" of the camera (1 or 2, default = 1)

        ------------------------------------------------------------------- '''
        shmf = name.lower() + ".im.shm"
        tmp = Cam(name=name, csz=self.csz, ysz=ysz, xsz=xsz,
                  pupil=self.tel.pupil, pdiam=self.tel.pdiam,
                  pscale=pscale, wl=wl, shdir=self.shdir, shmf=shmf)

        self._install_camera(tmp, slot=slot)

    # ==================================================
    def add_corono_camera(self, name="corocam", ysz=256, xsz=320,
                          lstop=None, fpm=None, pview=False,
                          pscale=16.7, wl=1.6e-6, slot=1):
        ''' -------------------------------------------------------------------
        Adds a coronagraphic camera to the instrument.

        Parameters:
        ----------
        - name   : string describing the camera (used for shm file name)
        - ysz    : vertical size of the detector in pixels (default = 256)
        - xsz    : horizontal size of the detector in pixels (default = 320)
        - lstop  : 2D array describing the Lyot-stop (default = None)
        - fpm    : 2D array describing the focal plane mask (default = None)
        - pview  : is this a pupil viewing camera, like ZELDA (default = False)
        - pscale : detector plate scale in mas/pixels (default = 16.7)
        - wl     : wavelength of observation in meters (default = 1.6e-6)
        - slot   : marks the "position" of the camera (1 or 2, default = 1)

        ------------------------------------------------------------------- '''
        shmf = name.lower() + ".im.shm"
        tmp = CoroCam(name=name, csz=self.csz, ysz=ysz, xsz=xsz,
                      pupil=self.tel.pupil, pdiam=self.tel.pdiam,
                      fpm=fpm, lstop=lstop, pview=pview,
                      pscale=pscale, wl=wl, shdir=self.shdir, shmf=shmf)

        self._install_camera(tmp, slot=slot)

    # ==================================================
    def add_SH_camera(self, name="SHcam", dsz=240, mls=40,
                      wl=0.7e-6, slot=1):
        ''' -------------------------------------------------------------------
        Adds a Shack-Hartman WFS camera to the instrument.

        Parameters:
        ----------
        - name   : string describing the camera (used for shm file name)
        - dsz    : detector size (square) in pixels (default = 240)
        - mls    : micro-lens array size (integer, default = 40)
        - wl     : wavelength of operation (float, default = 0.7e-6)
        - slot   : marks the "position" of the camera (1 or 2, default = 1)

        Notes:
        -----
        I realize that there is no plate scale parameter here: the assumption
        is that the SH images are always just nyquist sampled... which may
        not always be desirable...
        ------------------------------------------------------------------- '''
        shmf = name.lower() + ".im.shm"
        tmp = SHCam(name=name, csz=self.csz, dsz=dsz, mls=mls,
                    pupil=self.tel.pupil, wl=wl, shdir=self.shdir, shmf=shmf)

        self._install_camera(tmp, slot=slot)

    # ==================================================
    def add_membrane_DM(
            self, dms=50, nch=4, na0=49, iftype="cosine",
            ifr0=1.0, dtype=np.float64):
        ''' -------------------------------------------------------------------
        Adds a continuous membrane DM made of a regular grid of actuators
        to an instrument.

        Parameters:
        ----------
        - dms    : linear size of the DM in actuators (default = 50)
        - nch    : number of virtual DM channels created (default = 4)
        - na0    : number of actuators across a pupil diameter (default = 49)
        - iftype : influence function type (default = "cosine"
        - ifr0   : influence function radius in actuator size (default = 1.0)
        - dtype  : the data type written to shared memory (default np.float64)
        ------------------------------------------------------------------- '''
        if self.DM is None:
            self.DM = DM(dms=dms, nch=nch, shdir=self.shdir,
                         csz=self.csz, na0=na0, iftype=iftype, dtype=dtype)
        else:
            print("A DM is already in place. Remove to replace it!")

    # ==================================================
    def add_hex_DM(self, nr=7, nch=4, na0=15, srad=323.75):
        ''' -------------------------------------------------------------------
        Adds a segmented mirror with hexagonal geometry to an instrument.

        Parameters:
        ----------
        - nr   : number of rings of segments (default = 7)
        - nch  : number of virtual DM channels created (default = 4)
        - na0  : number of segments across a pupil diameter (default = 15)
        - srad : physical radius of a segment in microns (default = 323.75)

        Note:
        ----

        srad only matters when applying tip-tilt to segments, to convert the
        tip-tilt values (in mrad) into actuator vertical displacements.
        Assumed actuator layout behind segments is that of the BMC Hex-507.
        ------------------------------------------------------------------- '''
        if self.DM is None:
            self.DM = HexDM(nr=nr, nch=nch, csz=self.csz, na0=na0, srad=srad)
        else:
            print("A DM is already in place. Remove to replace it!")

    # ==================================================
    def add_phscreen(self, name="MaunaKea", r0=0.5, L0=10.0, wl=0.5e-6,
                     fc=24.5, correc=10.0):
        ''' -------------------------------------------------------------------
        Adds an atmospheric phase screen above the instrument.
        Phase structure is Von-Karman with partial correction by some factor
        up to a cut-off spatial frequency.

        Parameters:
        ----------
        - name   : string describing the phase screen (used by shm file name)
        - r0     : Fried parameter in meters (default = 0.5)
        - wl     : wavelength where r0 is specified (in meters)
        - L0     : outer scale parameter in meters (default = 10)
        - fc     : cut-off frequency of correction in l/D (default = 24.5)
        - correc : uniform correction factor up until fc (default = 10)

        Note:
        ----
        correc = 10 means that up to spatial frequency fc, the PSD of the
        phase screen is attenuated by a factor 10.
        ------------------------------------------------------------------- '''
        shmf = name.lower() + ".im.shm"
        if self.atmo is None:
            self.atmo = Phscreen(name=name, csz=self.csz, lsz=self.tel.pdiam,
                                 r0=r0, wl=wl, L0=L0, fc=fc, correc=correc,
                                 shdir=self.shdir, shmf=shmf)
        else:
            print("""An atmospheric phase screen is already in place.
            You can update its properties or remove it from the simulation""")

    # ==================================================
    def clear_DM(self):
        ''' -------------------------------------------------------------------
        Removes the DM from the instrument.

        Removal is required prior to installing a new DM
        ------------------------------------------------------------------- '''
        try:
            self.DM.close()
            del self.DM
            self.DM = None
        except AttributeError:
            print("No DM to remove from instrument")

    # ==================================================
    def clear_atmo(self):
        ''' -------------------------------------------------------------------
        Removes the atmospheric phase screen from the instrument.
        ------------------------------------------------------------------- '''
        try:
            _ = self.atmo.shm_phs.fd
            if (self.atmo.shm_phs.fd != 0):
                self.atmo.shm_phs.close()
                del self.atmo
                self.atmo = None

        except AttributeError:
            print("No atmo to remove from instrument")

    # ==================================================
    def snap(self):
        ''' -------------------------------------------------------------------
        Image snap produced without starting servers.

        Mode useful for simulations not requiring the semi real-time
        features of xaosim.

        Returns the image acquired by camera on slot #1 but computes
        images by multiple cameras if relevant.
        ------------------------------------------------------------------- '''

        dmmap = None
        phscreen = None

        if self.DM is not None:
            self.DM.update(verbose=False)
            dmmap = self.DM.wft.get_data()

        if self.atmo is not None:
            phscreen = self.atmo.shm_phs.get_data()

        if self.cam is not None:
            self.cam.make_image(dmmap=dmmap, opdmap=phscreen)
        if self.cam2 is not None:
            self.cam2.make_image(dmmap=dmmap, opdmap=phscreen)
        return(self.cam.get_image())

    # ==================================================
    def start(self, delay=0.1):
        ''' -------------------------------------------------------------------
        A function that starts all the components *servers*

        To each component is associated a server that periodically updates
        information on the global DM shape and the camera image, based on
        the status of the atmospheric phase screen and the different DM
        channels.
        -------------------------------------------------------------------
        Parameter:
        ---------
        - delay: (float) a time delay in seconds that sets a common cadence

        Usage:
        -----

        >> myinstrument.start(delay=0.2)

        ------------------------------------------------------------------- '''

        self.delay = delay

        dm_shmf = None
        atmo_shmf = None

        if self.DM is not None:
            self.DM.start(delay)
            dm_shmf = self.shdir+self.DM.shmf

        if self.atmo is not None:
            self.atmo.start(delay)
            atmo_shmf = self.shdir+self.atmo.shmf

        for dev in [self.cam, self.cam2]:
            if dev is not None:
                dev.start(delay=delay, dm_shmf=dm_shmf, atmo_shmf=atmo_shmf)
                del dev

    # ==================================================
    def stop(self,):
        ''' -------------------------------------------------------------------
        Stops all the threads of the different devices part of the instrument.
        ------------------------------------------------------------------- '''
        for dev in [self.atmo, self.DM, self.cam, self.cam2]:
            try:
                _ = dev
                if dev is not None:
                    dev.stop()
                    del dev
            except AttributeError:
                pass

    # ==================================================
    def __del__(self,):
        self.close()
        print("instrument was effectively destroyed!")

    # ==================================================
    def close(self,):
        ''' -------------------------------------------------------------------
        Closes shared memory data structures and release file descriptors.
        ------------------------------------------------------------------- '''
        self.stop()  # just in case!
        time.sleep(self.delay)

        self.clear_atmo()
        self.clear_DM()
        self.clear_camera_slot(slot=1)
        self.clear_camera_slot(slot=2)

        try:
            _ = self.tel
            del self.tel
        except AttributeError:
            pass



import numpy as np
from .pupil import uniform_rect as urect
from .pupil import _xyic
from scipy.ndimage import rotate

shift = np.fft.fftshift
fft = np.fft.fft2
ifft = np.fft.ifft2


# ======================================================================
def sin_map(sz, ncx, ncy, phi=0.0):
    ''' -----------------------------------------------------------
    Produces the 2D map of a sinusoidal modulation with a given
    number of cycles across the aperture.

    Parameters:
    ----------
    - sz : the size of the map in pixels (sz x sz)
    - ncx: the horizontal number of cycles
    - ncy: the vertical number of cycles
    - phi: the phase of the modulation (in radians)

    Remarks:
    -------
    phi = 0       means a true sinusoidal modulation.
    phi = np.pi/2 means cosinusoidal
    ----------------------------------------------------------- '''
    ramp = np.arange(sz)/(sz)-0.5
    xx, yy = np.meshgrid(ramp, ramp)
    res = np.sin(2*np.pi*(xx*ncx + yy*ncy) + phi)
    return res.astype(np.float32)


# ======================================================================
def mksin_vector(xymask, ncx, ncy, phi=0.0):
    ''' -----------------------------------------------------------
    returns a 1D vector of size xymask.shape[0], containing the
    local values of a sinusoidal wavefront computed for the xymask
    set of coordinates.
    ----------------------------------------------------------- '''
    xmax = np.abs(xymask[:, 0]).max()
    ymax = np.abs(xymask[:, 1]).max()
    fcx = ncx / xmax
    fcy = ncy / ymax
    res = np.sin(np.pi*(fcx * xymask[:, 0] + fcy * xymask[:, 1]) + phi)
    return res


# ======================================================================
def poke_map(sz, dx, dy, ww=3):
    ''' -----------------------------------------------------------
    Produces a 2D poke map for the aperture at the position dx, dy

    Parameters:
    ----------
    - sz: the size of the map in pixels (sz x sz)
    - dx: the horizontal offset of the poke (in pixels)
    - dy: the vertical offset of the poke (in pixels)
    - ww: the width of the poke (in pixels)

    Remarks:
    -------
    To add a border width eliminates the risk for a poke
    applied on one edge of the aperture to also appear on the
    opposite edge
    ----------------------------------------------------------- '''
    bw = 2*ww
    poke = urect(sz+2*bw, sz+2*bw, ww, ww, between_pix=False)
    poke = np.roll(np.roll(poke, dy, axis=1), dx, axis=0)
    return poke[bw:bw+sz, bw:bw+sz]


# ======================================================================
def piston_map(sz, coords, hrad, between_pix=True, piston=None):
    ''' -----------------------------------------------------------
    Produces a 2D map with circular holes where a piston is applied

    Parameters:
    ----------
    - sz          : the size of the map in pixels (sz x sz)
    - coords      : the array of hole coordinates (in pixels)
    - hrad        : the hole radius (in pixels)
    - between_pix : to align the map between pixels (bool)
    - piston      : a 1D array (or list) of pistons

    Remarks:
    -------
    If different than None, "piston" should contain at least as
    many values as there are holes. If None, a random set of
    values is genreated.
    ----------------------------------------------------------- '''

    off = 0
    if between_pix is True:
        off = 0.5
    xx, yy = np.meshgrid(np.arange(sz)-sz/2+off, np.arange(sz)-sz/2+off)
    mydist = np.hypot(yy, xx)
    pmap = np.zeros((sz, sz))

    p0 = np.zeros(coords.shape[0])
    if piston is not None:
        p0 = piston
    else:
        p0 = np.random.randn(coords.shape[0])

    xs = coords[:, 0]
    ys = coords[:, 1]
    for i in range(xs.size):
        pmap = np.roll(np.roll(pmap, -ys[i], 0), -xs[i], 1)
        pmap[mydist < hrad] = p0[i]
        pmap = np.roll(np.roll(pmap,  ys[i], 0),  xs[i], 1)
    return pmap


# ======================================================================
def kolmo(rnd, fc, ld0, correc=1e0, rms=0.1):
    ''' -----------------------------------------------------------
    WARNING: "legacy" function kept here as a reference.

    Use atmo_screen() instead!

    This function was adapted to a FFT based diffraction simulation,
    which is no longer used by this package.

    Does a Kolmogorov wavefront simulation with partial AO correction.

    Wavefront simulation of total size "size", following Kolmogorov
    statistics with a Fried parameter "r0", with partial AO correction
    up to a cutoff frequency "fc".

    Parameters:
    ----------

    - rnd     : array of uniformly distributed numbers [0,1)
    - fc      : cutoff frequency (in lambda/D)
    - ld0     : lambda/D (in pixels)
    - correc  : correction of wavefront amplitude (factor 10, 100, ...)
    - rms     : rms over the entire computed wavefront

    Note1: after applying the pupil mask, the Strehl is going to vary a bit
    Note2: one provides rnd from outside the routine so that the same
    experiment can be repeated with the same random numbers.
    ----------------------------------------------------------- '''

    ys, xs = rnd.shape
    xx, yy = np.meshgrid(np.arange(xs) - xs/2, np.arange(ys) - ys/2)
    rr = shift(np.hypot(yy, xx))

    in_fc = (rr < (fc*ld0))

    rr[0, 0] = 1.0  # trick to avoid div by 0!
    modul = rr**(-11./6.)
    modul[in_fc] /= correc

    test = (ifft(modul * np.exp(1j * 2*np.pi * (rnd - 0.5)))).real

    test -= np.mean(test)
    test *= rms/np.std(test)

    return test


# ==================================================================
def atmo_screen(isz, ll, r0, L0, fc=25, correc=1.0, pdiam=None):
    ''' -----------------------------------------------------------
    The Kolmogorov - Von Karman phase screen generation algorithm.

    Adapted from the work of Carbillet & Riccardi (2010).
    http://cdsads.u-strasbg.fr/abs/2010ApOpt..49G..47C

    Kolmogorov screen can be altered by an attenuation of the power
    by a correction factor *correc* up to a cut-off frequency *fc*
    expressed in number of cycles across the phase screen

    Parameters:
    ----------

    - isz    : the size of the array to be computed (in pixels)
    - ll     :  the physical extent of the phase screen (in meters)
    - r0     : the Fried parameter, measured at a given wavelength (in meters)
    - L0     : the outer scale parameter (in meters)
    - fc     : DM cutoff frequency (in lambda/D)
    - correc : correction of wavefront amplitude (factor 10, 100, ...)
    - pdiam  : pupil diameter (in meters)

    Returns: two independent phase screens, available in the real and
    imaginary part of the returned array.

    Remarks:
    -------
    If pdiam is not specified, the code assumes that the diameter of
    the pupil is equal to the extent of the phase screen "ll".
    ----------------------------------------------------------- '''

    phs = 2*np.pi * (np.random.rand(isz, isz) - 0.5)

    xx, yy = np.meshgrid(np.arange(isz)-isz/2, np.arange(isz)-isz/2)
    rr = np.hypot(yy, xx)
    rr = shift(rr)
    rr[0, 0] = 1.0

    modul = (rr**2 + (ll/L0)**2)**(-11/12.)

    if pdiam is not None:
        in_fc = (rr < fc * ll / pdiam)
    else:
        in_fc = (rr < fc)

    modul[in_fc] /= correc

    screen = ifft(modul * np.exp(1j*phs)) * isz**2
    screen *= np.sqrt(2*0.0228)*(ll/r0)**(5/6.)

    screen -= screen.mean()
    return(screen)


# ==================================================================
def noll_variance(iz, D, r0):
    '''Computes the Noll residual variance (in rad**2) for a partly
    corrected wavefront.

    Adapted from the work of Marcel Carbillet for CAOS Library 5.4

    Itself based on (Noll R.J., JOSA, 66, 3 (1976))

    Parameters:
    ----------
    - iz : Zernike mode until which the wavefront is ideally corrected
    - D  : aperture diameter (in meters)
    - r0 : Fried parameter (in meters)
    -------------------------------------------------------------------
    '''
    noll_var = [1.0299, 0.582,  0.134, 0.111, 0.0880,
                0.0648, 0.0587, 0.0525, 0.0463, 0.0401,
                0.0377, 0.0352, 0.0328, 0.0304, 0.0279,
                0.0267, 0.0255, 0.0243, 0.0232, 0.022,
                0.0208]

    try:
        noll = noll_var[iz]
    except IndexError:
        noll = 0.2944*(iz+1.)**(-np.sqrt(3)/2.)

    return(noll*(D/r0)**(5./3))


# ==================================================================
def noll_rms(iz, D, r0, wl=None):
    '''Computes the RMS for a partly corrected wavefront.

    Adapted from the work of Marcel Carbillet for CAOS Library 5.4
    Itself based on (Noll R.J., JOSA, 66, 3 (1976))

    Uses the Noll residual variance "noll_variance()" function
    introduced above.

    Parameters:
    ----------
    - iz : Zernike mode until which the wavefront is ideally corrected
    - D  : aperture diameter (in meters)
    - r0 : Fried parameter (in meters)

    Optional parameter:
    ------------------
    - wl: the wavelength (unit TBD by user)

    If wl is provided, the result (otherwise in radians) is converted
    into OPD, expressed in matching units
    -------------------------------------------------------------------
    '''
    vari = noll_variance(iz, D, r0)
    if wl is None:
        res = np.sqrt(vari)
    else:
        res = np.sqrt(vari) * wl / (2*np.pi)
    return(res)


# =============================================================================
def sector_modes(sz, ns=4, ponly=False, rot=0.0):
    ''' ---------------------------------------------------------------
    Returns a cube of 2D sector wavefront modes

    Intended to simulate and/or fit LWE (low wind effect) like modes
    on a telescope. Going counterclockwise, from horizontal right

    Parameters:
    - sz    : size (integer) of the array returned (sz x sz)
    - ns    : number (integer) of sectors (ns=4 or 6)
    - ponly : piston only or piston + tip-tilt (default)
    - rot   : rotation angle (in degrees)
    --------------------------------------------------------------- '''
    nm = ns if ponly else ns * 3
    res = np.empty((nm, sz, sz))
    res[:] = np.nan
    yy, xx = _xyic(sz, sz)
    dist = np.hypot(yy, xx)

    th = np.linspace(0, 2*np.pi, ns, endpoint=False)

    sector = (xx >= 0) * (yy >= 0.0) * np.abs(np.arctan(yy/(xx+1e-8)) < th[1])
    sector = rotate(sector, -rot, order=0, reshape=False)
    sector[dist > sz // 2] = False
    if ponly:
        for ii in range(ns):
            res[ii] = rotate(sector, -360/ns*ii, order=0, reshape=False)
    else:
        for ii in range(ns):
            tmp = rotate(sector, -360/ns*ii, order=0, reshape=False)
            res[3*ii] = tmp
            res[3*ii+1] = xx * tmp
            res[3*ii+1, tmp] -= res[3*ii+1, tmp].mean()
            res[3*ii+2] = yy * tmp
            res[3*ii+2, tmp] -= res[3*ii+2, tmp].mean()
    for ii in range(nm):
        res[ii] *= 1.0/res[ii].std()
    return res



import numpy as np
from scipy.special import factorial as fac
from scipy import *


def dist(n, m):
    ''' -----------------------------------------------------------------------
    Returns the distance (in pixels) relative to the center of the array

    Parameters:
    ----------
    - (n,m): size of the array in pixels

    Remarks:
    -------
    - for odd size, center of array = central pixel
    - for even size, center of array is in between pixels!

    - equivalent of the convenient IDL dist function!
    ----------------------------------------------------------------------- '''
    if (n % 2 == 0):
        xl = np.arange(n) - n//2 + 0.5
    else:
        xl = np.arange(n) - n//2

    if (m % 2 == 0):
        yl = np.arange(m) - m//2 + 0.5
    else:
        yl = np.arange(m) - m//2

    xx, yy = np.meshgrid(xl, yl)
    return(np.hypot(yy, xx))


def azim(n, m):
    ''' -----------------------------------------------------------------------
    Returns the azimuth in radians of points in an array of size (n, m) with
    respect to the center of the array.

    Parameters:
    ----------
    - (n,m): size of the array in pixels

    Remarks:
    -------
    - for odd size, center of array = central pixel
    - for even size, center of array is in between pixels!

    - equivalent of the convenient IDL dist function!
    ----------------------------------------------------------------------- '''
    if (n % 2 == 0):
        xl = np.arange(n) - n//2 + 0.5
    else:
        xl = np.arange(n) - n//2

    if (m % 2 == 0):
        yl = np.arange(m) - m//2 + 0.5
    else:
        yl = np.arange(m) - m//2

    xx, yy = np.meshgrid(xl, yl)
    return np.arctan2(xx, yy)


def zer_coeff(n, m):
    ''' -----------------------------------------------------------------------
    Returns the Zernike coefficients and exponents for a given mode (n,m)
    ----------------------------------------------------------------------- '''
    coeffs, pows = [], []

    for s in range((n-m)//2+1):
        coeffs.append((-1.0)**s *
                      fac(n-s) /
                      (fac(s) * fac((n+m)/2.0 - s) * fac((n-m)/2.0 - s)))
        pows.append(n-2.0*s)
    return coeffs, pows


def noll_2_zern(j):
    '''------------------------------------------
    Noll index converted to Zernike indices

    j: Noll index
    n: radial Zernike index
    m: azimuthal Zernike index
   ------------------------------------------ '''
    if (j == 0):
        raise ValueError("Noll indices start at 1")

    n = 0
    j1 = j-1
    while (j1 > n):
        n += 1
        j1 -= n

    m = (-1)**j * ((n % 2) + 2 * int((j1+((n+1) % 2)) / 2.0))
    return (n, m)


def mkzer(n, m, size, rad, limit=False):
    '''----------------------------------------------------
    returns a 2D array of size sz,
    containing the n,m Zernike polynomial
    within a disk of radius rad

    Parameters:
    ----------
    - n, m: the Zernike indices for the polynomial
    - size: the size of the 2D array
    - rad: the radius of the circular area covered
    - limit: boolean (if True, mode extends beyond "rad"
   ---------------------------------------------------- '''
    res = np.zeros((size, size))
    (coeffs, pows) = zer_coeff(n, np.abs(m))

    rho = dist(size, size)/float(rad)
    outp = np.where(rho > 1.0)
    inp = np.where(rho <= 1.0)

    for i in range(np.size(coeffs)):
        res += coeffs[i] * (rho)**pows[i]

    if (limit is not False):
        res[outp] = 0.0

    azi = azim(size, size)

    if m > 0:
        res *= np.cos(m * azi)
    if m < 0:
        res *= np.sin(-m * azi)

    # normalization
    rms0 = np.std(res[inp])
    res /= rms0
    return res


def zer_mode_bank_2D(sz, i0, i1, limit=True):
    ''' ------------------------------------------
    Returns a 3D array containing 2D (sz x sz)
    maps of Zernike modes for Noll index going
    from i0 to i1 included.

    Parameters:
    ----------
    - sz: the size of the contained 2D arrays
    - i0: the first Zernike index to be used
    - i1: the last Zernike index to be used
    - limit: boolean (if True, mode confined to "rad")
    ------------------------------------------ '''
    dZ = i1 - i0 + 1
    res = np.zeros((dZ, sz, sz))
    for i in range(i0, i1+1):
        res[i-i0] = mkzer1(i, sz, sz/2, limit=limit)
    return(res)


def zer_name_list(i0, i1):
    ''' ------------------------------------------
    Returns a list of names of zernike modes for
    the Noll index going from i0 to i1 included.

    Note that the first Noll index is 1.
    ------------------------------------------ '''
    if not i1 > i0 > 0:
        return []
    zer_names = ['piston', 'x-tilt', 'y-tilt', 'defocus',
                 'obliq. astig.', 'vertc. astig.',
                 'vertc. coma', 'horiz. coma',
                 'vertc. trefoil', 'obliq. trefoil',
                 'prim. spherical',
                 'vertc. sec. astig.', 'obliq. sec. astig.',
                 'vertc. quadrafoil', 'obliq. quadrafoil']

    zer_list = [f"zernike_{ii:02d}" for ii in range(20)]
    for ii, lbl in enumerate(zer_names):
        zer_list[ii] = lbl
    return zer_list[i0-1:i1]


def mkzer1(jj, sz, rad, limit=False):
    ''' ----------------------------------------------------
    returns a 2D array of size sz,
    containing the j^th Zernike polynomial
    within a disk of radius rad

    Parameters:
    ----------
    - jj: the Noll index for the polynomial
    - sz: the size of the 2D array
    - rad: the radius of the circular area covered
    - limit: boolean (if True, mode confined to "rad")
   ---------------------------------------------------- '''
    (n, m) = noll_2_zern(jj)
    if jj == 1:
        return np.ones((sz, sz))
    else:
        return mkzer(n, m, sz, rad, limit)  # .astype(np.float32)


def mkzer_vector(n, m, xymask):
    '''------------------------------------------
    returns a 1D vector of size xymask.shape(0),
    containing the n,m Zernike polynomial
   ------------------------------------------ '''
    (coeffs, pows) = zer_coeff(n, np.abs(m))
    res = np.zeros(xymask.shape[0])
    rho = np.sqrt(xymask[:, 0]**2+xymask[:, 1]**2)
    rho /= rho.max()

    for i in range(np.size(coeffs)):
        res += coeffs[i] * (rho)**pows[i]

    azi = np.arctan2(xymask[:, 0], xymask[:, 1])

    if m > 0:
        res *= np.cos(m * azi)
    if m < 0:
        res *= np.sin(-m * azi)

    # normalization
    rms0 = np.std(res)
    res /= rms0
    return res


def mkzer1_vector(j, xymask):
    '''------------------------------------------
    returns a 1D vector of size xymask.shape(0),
    containing the j^th Zernike polynomial
   ------------------------------------------ '''
    (n, m) = noll_2_zern(j)
    return(mkzer_vector(n, m, xymask))


def mk_pattern(n, m):
    '''------------------------------------------
    returns an ad-hoc pattern which I do not
    remember what it corresponds to. Probably
    should discard this function.
   ------------------------------------------ '''
    x, y = np.meshgrid(np.arange(n)-n/2, np.arange(m)-m/2)
    dd = np.roll(np.hypot(y, x), 8, axis=0)
    b = np.zeros_like(dd)
    b[dd < 12] = 1.0
    b[dd < 7] = 0.0
    b[x < 0] = 0.0
    b[10:45, 20:25] = 1.0
    b[10:15, 17:28] = 1.0
    return(b)
